import ipaddress
from typing import Dict, List, Optional

from pydantic import BaseModel, validator

from common_osint_model.models import ShodanDataHandler, CensysDataHandler, Logger


class AutonomousSystem(BaseModel, ShodanDataHandler, CensysDataHandler, Logger):
    """Represents an autonomous system"""
    number: Optional[int]
    name: Optional[str]
    country: Optional[str]
    prefix: Optional[str]
    source: str

    @validator("prefix")
    def validate_prefix(cls, v):
        if not v:
            return v
        try:
            ipaddress.ip_network(v)
        except Exception as e:
            raise ValueError(f"Prefix given could not be parsed by ipaddress module. Likely \"{v}\" has a "
                             f"wrong format: {e}")
        return v

    @classmethod
    def from_shodan(cls, d: Dict):
        """Creates an instance of this class using a typical Shodan dictionary."""
        if isinstance(d, List):
            cls.debug("Got a list instead of a dictionary. Usually multiple services of the same host are represented "
                      "as multiple list items by shodan, so this should not be a problem as the AS is the same for all."
                      " Using the first item.")
            d = d[0]
        asn = d.get("asn", None)
        return AutonomousSystem(
            number=int(asn.replace("AS", "")) if asn and isinstance(asn, str) else None,
            name=d.get("isp"),
            country=d.get("location", {}).get("country_code", None),
            prefix=None,  # Not available in Shodan data
            source="shodan"
        )

    @classmethod
    def from_censys(cls, d: Dict):
        autonomous_system = d.get("autonomous_system", {})
        return AutonomousSystem(
            number=autonomous_system.get("asn", None),
            name=autonomous_system.get("name", None),
            country=autonomous_system.get("country_code", None),
            prefix=autonomous_system.get("bgp_prefix", None),
            source="censys"
        )



import hashlib
from binascii import hexlify
from typing import Tuple

import mmh3


def flatten(d: dict, parent_key: str = "") -> dict:
    """
    Flattens a dictonary so the attributes are accessible like the results of the censys api
    :param d: dictionary
    :param parent_key: parent key (used for recursion)
    :return: flattened dict (of level x, in recursion)
    """
    items = []
    for key, value in d.items():
        combined_key = "{}.{}".format(parent_key, key) if parent_key != "" else key
        try:
            items.extend(flatten(value, combined_key).items())
        except AttributeError:
            items.append((combined_key, value))
    return dict(items)


def unflatten(flattened: dict) -> dict:
    """
    Unflattens a dictionary
    :param flattened: Flattened dictionary
    :return: Unflattened dictionary
    """
    unflattened = {}
    for key, value in flattened.items():
        parts = key.split(".")
        d = unflattened
        for part in parts[:-1]:
            if part not in d:
                d[part] = dict()
            d = d[part]
        d[parts[-1]] = value
    return unflattened


def common_model_cn_extraction(g: dict) -> list:
    """
    Loops through all keys in the already converted model in order to find domains
    :param g: dictionary of generic model
    :return: list of domain objects
    """
    domains = []
    for key, value in g.items():
        if not isinstance(value, dict):
            continue

        if "tls" in value.keys():
            cns = value["tls"].get("certificate", {}).get("subject", {}).get("common_name", None) or list()
            for cn in cns:
                # Sloppy check if this is a real domain
                if "." in cn:
                    domains.append({"type": "common_name", "value": cn})
    return domains


def sha256_from_body_string(b: str) -> str:
    """
    Returns the sha256 hash of an html body given as string
    :param b: html body as string
    :return: hex digest of sha256 hash
    """
    h = hashlib.sha256()
    h.update(b.encode("utf8"))
    return hexlify(h.digest()).decode("ascii")


def list_cleanup(d: dict) -> dict:
    for k, v in d.items():
        if isinstance(v, dict):
            d[k] = list_cleanup(v)
        elif isinstance(v, list):
            if len(v) == 1:
                d[k] = v[0]
    return d


def hash_all(data: bytes) -> Tuple[str, str, str, str]:
    """
    Helper function to create all hashes for data given.

    :returns: Tuple of hashes as string: md5, sha1, sha256, murmur
    """
    md5, sha1, sha256, murmur = hashlib.md5(), hashlib.sha1(), hashlib.sha256(), mmh3.hash(data)
    md5.update(data), sha1.update(data), sha256.update(data)
    return md5.hexdigest(), sha1.hexdigest(), sha256.hexdigest(), murmur



from datetime import datetime
from typing import Dict, List, Optional

from pydantic import BaseModel

from common_osint_model.models import ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger
from common_osint_model.models.http import HTTPComponent
from common_osint_model.models.ssh import SSHComponent
from common_osint_model.models.tls import TLSComponent
from common_osint_model.models.dns import DNSComponent
from common_osint_model.utils import hash_all


class Service(BaseModel, ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger):
    """Represents a single service answering connections on specific ports."""
    port: int
    # Banner is optional as not every scanning service offers complete banners as response. Banners might be
    # reconstructed from the data, but some attributes might have the wrong order then (e.g. HTTP headers).
    # The according hashes are also not reliable because of this.
    banner: Optional[str]
    md5: Optional[str]
    sha1: Optional[str]
    sha256: Optional[str]
    murmur: Optional[str]
    # Every service object should include these timestamps. "timestamp" can be used for tracking the observation
    # timestamp from scanning services (e.g. Shodan)
    first_seen: Optional[datetime] = datetime.utcnow()
    last_seen: Optional[datetime] = datetime.utcnow()
    timestamp: Optional[datetime]
    # We need to include every possible service component here. In order to not export empty dictionary keys, the class
    # object can be exported with dict(exclude_none=True), so e.g. empty tls keys are skipped.
    http: Optional[HTTPComponent]
    tls: Optional[TLSComponent]
    ssh: Optional[SSHComponent]
    dns: Optional[DNSComponent]
    # Typically hosts consist of different services which might be discovered by different scanning services, so
    # remarking which service was observed by which scanner might be a good idea.
    source: str

    @classmethod
    def from_shodan(cls, d: Dict):
        """Creates an instance of this class using a dictionary with typical shodan data."""
        if isinstance(d, List):
            cls.info("The dictionary given is a list. Typically this list represents multiple services. Iterate over "
                     "the list to create Service objects for every item available. "
                     "This method just uses the first item.")
            d = d[0]

        port = d["port"]
        sshobj = None
        if "ssh" in d:
            sshobj = SSHComponent.from_shodan(d)

        httpobj = None
        if "http" in d:
            httpobj = HTTPComponent.from_shodan(d)

        tlsobj = None
        if "ssl" in d:
            tlsobj = TLSComponent.from_shodan(d)

        dnsobj = None
        if "dns" in d:
            dnsobj = DNSComponent.from_shodan(d)

        banner = d["data"]
        md5, sha1, sha256, murmur = hash_all(banner.encode("utf-8"))

        return Service(
            port=port,
            banner=d["data"],
            md5=md5,
            sha1=sha1,
            sha256=sha256,
            murmur=murmur,
            ssh=sshobj,
            http=httpobj,
            tls=tlsobj,
            dns=dnsobj,
            timestamp=datetime.fromisoformat(d["timestamp"]),
            source="shodan"
        )

    @classmethod
    def from_censys(cls, d: Dict):
        """Creates an instance of this class using a dictionary with typical Censys data."""
        port = d["port"]
        banner = d.get("banner", None)
        md5, sha1, sha256, murmur = None, None, None, None
        if banner:
            md5, sha1, sha256, murmur = hash_all(banner.encode("utf-8"))

        httpobj = None
        if "http" in d:
            httpobj = HTTPComponent.from_censys(d)

        tlsobj = None
        if "tls" in d:
            tlsobj = TLSComponent.from_censys(d)

        sshobj = None
        if "ssh" in d:
            sshobj = SSHComponent.from_censys(d)

        dnsobj = None
        if "dns" in d:
            dnsobj = DNSComponent.from_censys(d)

        return Service(
            port=port,
            banner=banner,
            md5=md5,
            sha1=sha1,
            sha256=sha256,
            murmur=murmur,
            http=httpobj,
            tls=tlsobj,
            ssh=sshobj,
            dns=dnsobj,
            timestamp=datetime.fromisoformat(d["observed_at"][:-4]),
            source="censys"
        )

    @classmethod
    def from_binaryedge(cls, d: List):
        """Creates an instance of this class using a dictionary with typical BinaryEdge data. Contrary to the other
        scanning services, binaryedge provides multiple entries per port."""
        port = d[0]["target"]["port"]
        type_index: Dict[str, int] = {service["origin"]["type"]: idx for idx, service in enumerate(d)}

        httpobj = None
        if "webv2" in type_index:
            httpobj = HTTPComponent.from_binaryedge(d[type_index["webv2"]])

        tlsobj = None
        if "ssl-simple" in type_index:
            tlsobj = TLSComponent.from_binaryedge(d[type_index["ssl-simple"]])

        sshobj = None
        if "ssh" in type_index:
            sshobj = SSHComponent.from_binaryedge(d[type_index["ssh"]])

        banner = None
        md5, sha1, sha256, murmur = None, None, None, None
        if "service-simple" in type_index:
            banner = d[type_index["service-simple"]]["result"]["data"]["service"].get("banner", None)
        if banner:
            md5, sha1, sha256, murmur = hash_all(banner.encode("utf-8"))

        return Service(
            port=port,
            http=httpobj,
            tls=tlsobj,
            ssh=sshobj,
            banner=banner,
            md5=md5,
            sha1=sha1,
            sha256=sha256,
            murmur=murmur,
            source="binaryedge"
        )



import warnings
from typing import Dict, List, Any

from DateTime import DateTime
from mmh3 import hash as mmh3_hash

from common_osint_model.utils import sha256_from_body_string, flatten


def from_censys(raw: Dict) -> Dict:
    """
    Convert Censys data model to the common data model.

    :param raw: Censys Search 2.0 dictionary
    """
    warnings.warn("This function was deprecated in v0.4.0. and will be removed in v0.5.0", DeprecationWarning)
    common = {}
    common.update(
        censys_meta_extraction(raw)
    )

    for service in raw.get("services", []):
        common.update(censys_extract_service(service))
    return common


def from_censys_flattened(raw: Dict) -> Dict:
    warnings.warn("This function was deprecated in v0.4.0. and will be removed in v0.5.0", DeprecationWarning)
    return flatten(from_censys(raw))


def censys_meta_extraction(raw: Dict) -> Dict:
    """
    Returns all metadata.

    :param raw: Censys Search 2.0 dictionary
    """
    _as = raw.get("autonomous_system", {})
    return {
        "ip": raw.get("ip", "None"),
        "as": {
            "name": _as.get("name", "None"),
            "number": _as.get("asn", "None"),
            "description": _as.get("description", "None"),
            "location": _as.get("country_code", "None"),
            "prefix": _as.get("bgp_prefix", "None")
        },
        "location": {
            **raw.get("location", {})
        },
        "ports": [service["port"] for service in raw.get("services", [])]
    }


def censys_extract_service(service: Dict) -> Dict:
    """
    Extracts relevant information from a service object/dict.

    :param service: Censys Search 2.0 service dictionary
    """
    port = service["port"]
    timestamp = service.get("observed_at", None)
    if timestamp:
        timestamp = DateTime(timestamp)
    s_common = {
        "banner": service.get("banner", None),
        "timestamp": int(timestamp),
        "timestamp_readable": timestamp.ISO8601()
    }
    if "http" in service:
        s_common.update({"http": censys_extract_http_service(service)})
    if "tls" in service:
        s_common.update({"tls": censys_extract_tls_service(service)})
    return {
        port: s_common
    }


def censys_extract_http_service(service: Dict) -> Dict:
    """Extracts relevant http service fields.

    :param service: Censys Search 2.0 service dictionary
    """
    s_http = {}
    res = service.get("http", {}).get("response", None)
    if not res:
        return {}

    headers = res.get("headers", None)
    if headers:
        s_http["headers"] = {}
    for k, v in headers.items():
        if k == "_encoding":
            continue

        s_http["headers"][k.lower()] = v[0]
    s_http["content"] = {
        "html": res.get("body"),
        "hash": {
            "shodan": mmh3_hash(res.get("body", None) or ""),
            "sha256": sha256_from_body_string(res.get("body", None) or ""),
            "censys": res.get("body_hash", None)
        }
    }
    return s_http


def censys_extract_tls_service(service: Dict) -> Dict:
    """Extracts relevant tls service fields.

    :param service: Censys Search 2.0 service dictionary
    """
    s_tls = {}
    cert = service.get("tls", {}).get("certificates", {}).get("leaf_data", None)
    c_issuer = cert.get("issuer", None) or dict()
    c_subject = cert.get("subject", None) or dict()
    common_name = c_subject.get("common_name", [])
    common_name.extend(cert.get("names", []))
    if len(common_name) == 0:
        common_name = None
    else:
        common_name = sorted(list(set(common_name)))
    if not cert:
        return {}

    s_tls["certificate"] = {
        "issuer_dn": cert.get("issuer_dn", None),
        "subject_dn": cert.get("subject_dn", None),
        "issuer": {
            "common_name": _first_or_none(c_issuer["common_name"]),
            # MISSING! "country": _first_or_none(c_issuer["country"]),
            "locality": _first_or_none(c_issuer["locality"]),
            "province": _first_or_none(c_issuer["province"]),
            "organization": _first_or_none(c_issuer["organization"]),
            "organizational_unit": _first_or_none(c_issuer["organizational_unit"]),
            # MISSING! "email_address": _first_or_none(c_issuer["email_address"]),
        },
        "subject": {
            "common_name": common_name,
            # MISSING! "country": _first_or_none(c_issuer["country"]),
            "locality": _first_or_none(c_subject["locality"]),
            "province": _first_or_none(c_subject["province"]),
            "organization": _first_or_none(c_subject["organization"]),
            "organizational_unit": _first_or_none(c_subject["organizational_unit"]),
            # MISSING! "email_address": _first_or_none(c_subject["email_address"]),
        },
        "fingerprint": {
            "sha256": cert.get("fingerprint", None)
        }
    }
    return s_tls


def censys_extract_ssh(service: Dict) -> Dict:
    """
    Extracts relevant ssh service fields.

    :param service: Censys Search 2.0 service dictionary
    """
    s_ssh = {}


def _first_or_none(l: List) -> Any:
    """Returns first element of list or none, if list is empty."""
    if not l:
        return None
    if len(l) > 0:
        return l[0]
    return None



import base64
from typing import Dict, List, Optional, Union

import mmh3
from pydantic import BaseModel

from common_osint_model.models import ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger
from common_osint_model.utils import hash_all


class HTTPComponentContentFavicon(BaseModel, ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger):
    """Represents the favicon which might be included in HTTP components."""
    raw: Optional[str]
    md5: Optional[str]
    sha1: Optional[str]
    sha256: Optional[str]
    murmur: Optional[str]
    shodan_murmur: Optional[str]

    @classmethod
    def from_shodan(cls, d: Dict):
        """Creates an instance of this class based on Shodan data given as dictionary."""
        if not isinstance(d, Dict):
            raise TypeError(f"Method HTTPComponentContentFavicon.from_shodan expects parameter d to be a dictionary, "
                            f"but it was {type(d)}.")

        raw = d["http"]["favicon"]["data"]
        raw = base64.b64decode(raw)
        md5, sha1, sha256, murmur = hash_all(raw)
        shodan_murmur = mmh3.hash(d["http"]["favicon"]["data"])
        cls.info("Shodan's favicon hash only hashes the base64 encoded favicon, not the data itself. The hash can be "
                 "found as \"shodan_murmur\" in this instance. \"murmur\" and the other hashes are calculated based on "
                 "the raw data of the favicon.")
        return HTTPComponentContentFavicon(
            raw=d["http"]["favicon"]["data"],
            md5=md5,
            sha1=sha1,
            sha256=sha256,
            murmur=murmur,
            shodan_murmur=shodan_murmur
        )

    @classmethod
    def from_censys(cls, d: Dict):
        """Not supported by Censys right now."""
        return None

    @classmethod
    def from_binaryedge(cls, d: Union[Dict, List]):
        favicon = d["result"]["data"]["response"]["favicon"]["content"]
        favicon_bytes = base64.b64decode(favicon.encode("utf-8"))
        md5, sha1, sha256, murmur = hash_all(favicon_bytes)
        shodan_murmur = mmh3.hash(favicon.encode("utf-8"))
        return HTTPComponentContentFavicon(
            raw=favicon,
            md5=md5,
            sha1=sha1,
            sha256=sha256,
            murmur=murmur,
            shodan_murmur=shodan_murmur
        )


class HTTPComponentContentRobots(BaseModel, ShodanDataHandler, CensysDataHandler):
    """Represents the robots.txt file in webroots."""
    raw: Optional[str]
    md5: Optional[str]
    sha1: Optional[str]
    sha256: Optional[str]
    murmur: Optional[str]

    @classmethod
    def from_shodan(cls, d: Dict):
        """Creates an instance of this class based on Shodan data given as dictionary."""
        if not isinstance(d, Dict):
            raise TypeError(
                f"Method HTTPComponentContentRobots.from_shodan expects parameter d to be a dictionary, "
                f"but it was {type(d)}.")

        raw = d["http"]["robots"].encode("utf-8")
        md5, sha1, sha256, murmur = hash_all(raw)
        return HTTPComponentContentRobots(
            raw=raw,
            md5=md5,
            sha1=sha1,
            sha256=sha256,
            murmur=murmur
        )

    @classmethod
    def from_censys(cls, d: Dict):
        """Not supported by Censys right now."""
        return None


class HTTPComponentContentSecurity(BaseModel, ShodanDataHandler, CensysDataHandler):
    """Represents the security.txt file in webroots."""
    raw: Optional[str]
    md5: Optional[str]
    sha1: Optional[str]
    sha256: Optional[str]
    murmur: Optional[str]

    @classmethod
    def from_shodan(cls, d: Dict):
        """Creates an instance of this class based on Shodan data given as dictionary."""
        if not isinstance(d, Dict):
            raise TypeError(
                f"Method HTTPComponentContentRobots.from_shodan expects parameter d to be a dictionary, "
                f"but it was {type(d)}.")

        raw = d["http"]["securitytxt"].encode("utf-8")
        md5, sha1, sha256, murmur = hash_all(raw)
        return HTTPComponentContentRobots(
            raw=raw,
            md5=md5,
            sha1=sha1,
            sha256=sha256,
            murmur=murmur
        )

    @classmethod
    def from_censys(cls, d: Dict):
        """Not supported by Censys right now."""
        return None


class HTTPComponentContent(BaseModel, ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger):
    """Represents the content (body) of HTTP responses."""
    raw: Optional[str]
    length: Optional[int]
    md5: Optional[str]
    sha1: Optional[str]
    sha256: Optional[str]
    murmur: Optional[str]
    favicon: Optional[HTTPComponentContentFavicon]
    robots_txt: Optional[HTTPComponentContentRobots]
    security_txt: Optional[HTTPComponentContentSecurity]

    @classmethod
    def from_shodan(cls, d: Dict):
        """Creates an instance of this class based on Shodan data given as dictionary."""
        if not isinstance(d, Dict):
            raise TypeError(f"Method HTTPComponentContent.from_shodan expects parameter d to be a dictionary, "
                            f"but it was {type(d)}.")

        favicon = None
        if "favicon" in d["http"]:
            cls.debug("Favicon key found in Shodan data.")
            favicon = HTTPComponentContentFavicon.from_shodan(d)

        security_txt = None
        if d["http"]["securitytxt"]:
            cls.debug("Security.txt key found in Shodan data.")
            security_txt = HTTPComponentContentSecurity.from_shodan(d)

        robots_txt = None
        if d["http"]["robots"]:
            cls.debug("Robots.txt key found in Shodan data.")
            robots_txt = HTTPComponentContentRobots.from_shodan(d)

        raw = d["http"].get("html", "")
        if not raw:
            raw = ""

        raw = raw.encode("utf-8")

        md5, sha1, sha256, murmur = hash_all(raw)
        return HTTPComponentContent(
            raw=raw,
            length=len(raw),
            md5=md5,
            sha1=sha1,
            sha256=sha256,
            murmur=murmur,
            favicon=favicon,
            robots_txt=robots_txt,
            security_txt=security_txt
        )

    @classmethod
    def from_censys(cls, d: Dict):
        """Creates an instance of this class based on Censys (2.0) data given as dictionary."""
        http = d["http"]["response"]
        raw = http["body"] if http["body_size"] > 0 else ""
        md5, sha1, sha256, murmur = hash_all(raw.encode("utf-8"))
        return HTTPComponentContent(
            raw=raw,
            length=len(raw),
            md5=md5,
            sha1=sha1,
            sha256=sha256,
            murmur=murmur,
            favicon=HTTPComponentContentFavicon.from_censys(d),
            robots_txt=HTTPComponentContentRobots.from_censys(d),
            security_txt=HTTPComponentContentSecurity.from_censys(d)
        )

    @classmethod
    def from_binaryedge(cls, d: Union[Dict, List]):
        """Creates an instance of this class based on BinaryEdge data given as dictionary. Robots and Security.txt are
        not supported by BinaryEdge."""
        http_response = d["result"]["data"]["response"]
        raw = http_response["body"]["content"]
        md5, sha1, sha256, murmur = hash_all(raw.encode("utf-8"))
        return HTTPComponentContent(
            raw=raw,
            length=len(raw),
            md5=md5,
            sha1=sha1,
            sha256=sha256,
            murmur=murmur,
            favicon=HTTPComponentContentFavicon.from_binaryedge(d)
        )


class HTTPComponent(BaseModel, ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler):
    """Represents the HTTP component of services."""
    headers: Optional[Dict[str, str]]
    content: Optional[HTTPComponentContent]
    shodan_headers_hash: Optional[str]

    @classmethod
    def from_shodan(cls, d: Dict):
        """Creates an instance of this class based on Shodan data given as dictionary."""
        if not isinstance(d, Dict):
            raise TypeError(f"Method HTTPComponent.from_shodan expects parameter d to be a dictionary, "
                            f"but it was {type(d)}.")

        content = HTTPComponentContent.from_shodan(d)
        banner = d["data"]
        lines = banner.split("\r\n")
        headers = {}
        banner_keys = lines[0]
        for line in lines:
            if ":" in line:
                key, value = line.split(":", maxsplit=1)
                headers[key.strip()] = value.strip()

        return HTTPComponent(
            headers=headers,
            content=content,
            shodan_headers_hash=d.get("http", {}).get("headers_hash", None)
        )

    @classmethod
    def from_censys(cls, d: Dict):
        """Todo: Is parsing from services.banner better than just looping over the headers found by Censys?"""
        http = d["http"]["response"]
        headers = {}
        for k, v in http["headers"].items():
            if k[0] == "_":
                continue

            headers.update({
                k.replace("_", "-"): " ".join(v)
            })

        banner_lines = d["banner"].replace("\r", "").split("\n")
        banner_keys = banner_lines[0]
        for line in banner_lines:
            if ":" in line:
                k, _ = line.split(":", maxsplit=1)
                banner_keys += "\n" + k
        headers_hash = mmh3.hash(banner_keys.encode("utf-8"))

        return HTTPComponent(
            headers=headers,
            content=HTTPComponentContent.from_censys(d),
            shodan_headers_hash=headers_hash
        )

    @classmethod
    def from_binaryedge(cls, d: Union[Dict, List]):
        http_response = d["result"]["data"]["response"]
        headers = http_response["headers"]["headers"]
        return HTTPComponent(
            headers=headers,
            content=HTTPComponentContent.from_binaryedge(d)
        )



import binascii
from datetime import datetime
from typing import Dict, List, Optional, Union

import pytz
from cryptography.hazmat.primitives.hashes import MD5, SHA1, SHA256
from cryptography.x509 import OID_COMMON_NAME, ExtensionOID, DNSName, ExtensionNotFound
from cryptography.x509 import load_pem_x509_certificate
from pydantic import BaseModel

from common_osint_model.models import ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger


class TLSComponentCertificateEntity(BaseModel, ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler):
    """Represents certificate entities, typically issuer and subject."""
    dn: Optional[str]
    country: Optional[str]
    state: Optional[str]
    locality: Optional[str]
    organization: Optional[str]
    organizational_unit: Optional[str]
    common_name: Optional[str]
    email_address: Optional[str]

    @classmethod
    def from_shodan(cls, d: Dict):
        """Creates an instance of this class using a given Shodan data dictionary."""
        if all(key not in d for key in ["C", "L", "CN", "O", "ST"]):
            raise KeyError("The dictionary given to TLSComponentCertificateEntity.from_shodan is missing the typical "
                           "shodan keys.")

        c = d.get("C", None)
        st = d.get("ST", None)
        l = d.get("L", None)
        o = d.get("O", None)
        ou = d.get("OU", None)
        cn = d.get("CN", None)
        email = d.get("emailAddress", None)
        dn = ""
        if c:
            dn += f"C={c}, "
        if st:
            dn += f"ST={st}, "
        if l:
            dn += f"L={l}, "
        if o:
            dn += f"O={o}, "
        if ou:
            dn += f"OU={ou}, "
        if cn:
            if not email:
                dn += f"CN={cn}"
            else:
                dn += f"CN={cn}/Email={email}"
        elif not cn and email:
            dn += f"Email={email}"

        if len(dn) > 0:
            while dn[-1] in [",", " "]:
                dn = dn[:-1]

        return TLSComponentCertificateEntity(
            dn=dn,
            country=c,
            state=st,
            locality=l,
            organization=o,
            organizational_unit=ou,
            common_name=cn,
            email=email
        )

    @classmethod
    def from_censys(cls, d: Dict):
        """Creates an instance of this class based on Censys data given as dictionary."""
        if all(key not in d for key in ["common_name", "locality", "organization", "organizational_unit", "province"]):
            raise KeyError("The dictionary given to TLSComponentCertificateEntity.from_censys is missing the typical "
                           "censys keys.")

        c = d.get("country", [])
        st = d.get("province", [])
        l = d.get("locality", [])
        o = d.get("organization", [])
        ou = d.get("organizational_unit", [])
        cn = d.get("common_name", [])
        email = d.get("email_address", [])
        dn = ""
        if c:
            for item in c:
                dn += f"C={item}, "
        if st:
            for item in st:
                dn += f"ST={item}, "
        if l:
            for item in l:
                dn += f"L={item}, "
        if o:
            for item in o:
                dn += f"O={item}, "
        if ou:
            for item in ou:
                dn += f"OU={item}, "
        done = False
        if email and cn:
            if len(email) == 1 and len(cn) == 1:
                dn += f"CN={cn[0]}/Email={email[0]}"
                done = True
            else:
                for item in cn:
                    dn += f"CN={item}, "
                for item in email:
                    dn += f"Email={item}, "
                done = True
        if cn and not done:
            for item in cn:
                dn += f"CN={item}, "

        # This one is probably wrong.
        if email and not done:
            for item in email:
                dn += f"Email={item}, "

        while dn[-1] in [" ", ","]:
            dn = dn[:-1]
        return TLSComponentCertificateEntity(
            dn=dn,
            country=", ".join(c),
            state=", ".join(st),
            locality=", ".join(l),
            organization=", ".join(o),
            organizational_unit=", ".join(ou),
            common_name=", ".join(cn),
            email=", ".join(email)
        )

    @classmethod
    def from_binaryedge(cls, d: Union[Dict, List]):
        c = d.get("country_name", None)
        st = d.get("state_or_province_name", None)
        l = d.get("locality_name", None)
        o = d.get("organization_name", None)
        ou = d.get("organizational_unit_name", None)
        cn = d.get("common_name", None)
        email = d.get("email_address", None)  # Todo: Check if this key is actually correct

        dn = ""
        if c:
            dn += f"C={c}, "
        if st:
            dn += f"ST={st}, "
        if l:
            dn += f"L={l}, "
        if o:
            dn += f"O={o}, "
        if ou:
            dn += f"OU={ou}, "
        if cn:
            if not email:
                dn += f"CN={cn}"
            else:
                dn += f"CN={cn}/Email={email}"
        elif not cn and email:
            dn += f"Email={email}"

        while dn[-1] in [",", " "]:
            dn = dn[:-1]

        return TLSComponentCertificateEntity(
            dn=dn,
            country=c,
            state=st,
            locality=l,
            organization=o,
            organizational_unit=ou,
            common_name=cn,
            email=email
        )


class TLSComponentCertificate(BaseModel, ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger):
    """Represents certificates."""
    issuer: Optional[TLSComponentCertificateEntity]
    subject: Optional[TLSComponentCertificateEntity]
    issued: Optional[datetime]
    expires: Optional[datetime]
    expired: Optional[bool]
    # More specifically, this is a certificate extension, but we keep it here because it's easier this way.
    alternative_names: Optional[List[str]]
    # The certificate itself
    pem: Optional[str]
    md5: Optional[str]
    sha1: Optional[str]
    sha256: Optional[str]
    murmur: Optional[str]
    # If the certificate is trusted by the source
    trusted: Optional[bool]

    @property
    def domains(self) -> List[str]:
        domains = []
        if self.subject.common_name:
            domains.append(self.subject.common_name)
        if self.alternative_names:
            domains.extend(self.alternative_names)
        return list(set(domains))

    @classmethod
    def from_shodan(cls, d: Dict):
        """Creates an instance of this class based on Shodan data given as dictionary."""
        if not isinstance(d, Dict):
            raise TypeError(f"Method TLSComponentCertificate.from_shodan expects parameter d to be a dictionary, "
                            f"but it was {type(d)}.")

        issuer = TLSComponentCertificateEntity.from_shodan(d["ssl"]["cert"]["issuer"])
        subject = TLSComponentCertificateEntity.from_shodan(d["ssl"]["cert"]["subject"])
        issued = datetime.strptime(d["ssl"]["cert"]["issued"], "%Y%m%d%H%M%SZ")
        expires = datetime.strptime(d["ssl"]["cert"]["expires"], "%Y%m%d%H%M%SZ")
        expired = True if d["ssl"]["cert"]["expired"] in ["true", True] else False
        altnames = []
        pem = None
        cert = None
        md5, sha1, sha256 = None, None, None
        certificate_chain = d["ssl"]["chain"]

        trusted = True
        tags = d.get("tags", None)
        if tags:
            if "self-signed" in tags:
                trusted = False

        # If only a single certificate is given in the chain, use it. Also set trusted to false, as this is likely
        # self-signed.
        if len(certificate_chain) == 1:
            pem = certificate_chain[0]
            cert = load_pem_x509_certificate(pem.encode("utf-8"))
        # If there are multiple certificates given, we need to loop over them and compare the common name. This _can_
        # lead to ValueError if the certificates are malformed, such as empty Country values etc. This is why
        # >>> cert.subject.get_attributes_for_oid(OID_COMMON_NAME)
        # is in a try/except block. In these cases, no further data will be extracted from the certificate. However,
        # malformed certificates are often self-signed and the chain length is 1.
        else:
            for cert_pem in d["ssl"]["chain"]:
                cert = load_pem_x509_certificate(cert_pem.encode("utf-8"))
                # Check if this certificate is the leaf certificate by comparing the common name
                attributes = []
                try:
                    attributes = cert.subject.get_attributes_for_oid(OID_COMMON_NAME)
                except:
                    cls.info("Could not get attributes for OID_COMMON_NAME. Skipping this certificate.")
                    continue
                for attribute in attributes:
                    if attribute.value == subject.common_name:
                        pem = cert_pem

        if cert:
            md5, sha1, sha256 = (
                binascii.hexlify(cert.fingerprint(MD5())).decode("utf-8"),
                binascii.hexlify(cert.fingerprint(SHA1())).decode("utf-8"),
                binascii.hexlify(cert.fingerprint(SHA256())).decode("utf-8")
            )

            try:
                ext = cert.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
                altnames.extend(ext.value.get_values_for_type(DNSName))
            except ExtensionNotFound:
                cls.debug("Could not extract alternative names from the certificate extensions.")

        if len(altnames) == 0:
            altnames = None
        else:
            # This removes duplicates
            altnames = list(set(altnames))

        return TLSComponentCertificate(
            issuer=issuer,
            subject=subject,
            issued=issued,
            expires=expires,
            expired=expired,
            alternative_names=altnames,
            pem=pem,
            md5=md5,
            sha1=sha1,
            sha256=sha256,
            trusted=trusted
        )

    @classmethod
    def from_censys(cls, d: Dict):
        """Creates an instance of this class based on Censys data given as dictionary."""
        cls.info("Censys does not provide raw certificate data, to hashes must be taken from the data and cannot be "
                 "calculated.")
        trusted = not d.get("signature", {}).get("self_signed", False)
        return TLSComponentCertificate(
            issuer=TLSComponentCertificateEntity.from_censys(d["issuer"]),
            subject=TLSComponentCertificateEntity.from_censys(d["subject"]),
            issued=None,
            expires=None,
            expired=None,
            alternative_names=d.get("names", None),
            sha256=d["fingerprint"],
            trusted=trusted
        )

    @classmethod
    def from_binaryedge(cls, d: Union[Dict, List]):
        pem = d["as_pem"]
        data = d["as_dict"]
        cert = load_pem_x509_certificate(pem.encode("utf-8"))
        md5, sha1, sha256 = (
            binascii.hexlify(cert.fingerprint(MD5())).decode("utf-8"),
            binascii.hexlify(cert.fingerprint(SHA1())).decode("utf-8"),
            binascii.hexlify(cert.fingerprint(SHA256())).decode("utf-8")
        )
        issued = datetime.fromisoformat(data["validity"]["not_before"]).replace(tzinfo=pytz.utc)
        expires = datetime.fromisoformat(data["validity"]["not_after"]).replace(tzinfo=pytz.utc)
        expired = datetime.utcnow().replace(tzinfo=pytz.utc) < expires
        trusted = not data.get("self_issued", False) or data.get("self_signed", False)
        return TLSComponentCertificate(
            issuer=TLSComponentCertificateEntity.from_binaryedge(data["issuer"]),
            subject=TLSComponentCertificateEntity.from_binaryedge(data["subject"]),
            issued=issued,
            expires=expires,
            expired=expired,
            alternative_names=data["extensions"]["subject_alt_name"],
            pem=pem,
            md5=md5,
            sha1=sha1,
            sha256=sha256,
            trusted=trusted
        )


class TLSComponent(BaseModel, ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger):
    """Represents the TLS component of services."""
    certificate: Optional[TLSComponentCertificate]
    ja3: Optional[str]
    ja3s: Optional[str]
    jarm: Optional[str]

    # Todo: Add other attributes relevant to TLS such as CipherSuits etc.

    @classmethod
    def from_shodan(cls, d: Dict):
        """Creates an instance of this class based on Shodan data given as dictionary."""
        if not isinstance(d, Dict):
            raise TypeError(f"Method TLSComponent.from_shodan expects parameter d to be a dictionary, "
                            f"but it was {type(d)}.")

        return TLSComponent(
            certificate=TLSComponentCertificate.from_shodan(d),
            ja3s=d.get("ssl", {}).get("ja3s", None),
            jarm=d.get("ssl", {}).get("jarm", None)
        )

    @classmethod
    def from_censys(cls, d: Dict):
        try:
            tls = d["tls"]
            return TLSComponent(
                certificate=TLSComponentCertificate.from_censys(tls["certificates"]["leaf_data"]),
                ja3s=tls.get("ja3s", None),
                jarm=d.get("jarm", {}).get("fingerprint", None)
            )
        except KeyError as e:
            cls.error(f"Exception during certificate data extraction. "
                      f"The key 'tls.certificates.leaf_data' is not available: {e}\nReturning None...")
            return None

    @classmethod
    def from_binaryedge(cls, d: Union[Dict, List]):
        """Creates an instance of this class based on BinaryEdge data given as dictionary."""
        data = d.get("result", {}).get("data", None)
        if not data:
            cls.error("No data key available in binary edge dictionary. Returning None...")
            return None

        cert = None
        if "cert_info" in data:
            cert = TLSComponentCertificate.from_binaryedge(
                data["cert_info"]["certificate_chain"][0]
            )
        ja3 = None
        if "server_info" in data:
            ja3 = data["server_info"].get("ja3_digest", None)

        jarm = None
        if "jarm_hash" in data:
            jarm = data["jarm_hash"]

        return TLSComponent(
            certificate=cert,
            ja3=ja3,
            jarm=jarm
        )



from abc import ABC
from logging import getLogger, basicConfig
from typing import Dict, List, Union

basicConfig(level="INFO")


class Logger(ABC):
    """Abstract class which implements just an info method printing a message to stdout via Logger class."""

    @classmethod
    def info(cls, message: str):
        logger = getLogger(cls.__name__)
        logger.info(message)

    @classmethod
    def debug(cls, message: str):
        logger = getLogger(cls.__name__)
        logger.debug(message)

    @classmethod
    def warning(cls, message: str):
        logger = getLogger(cls.__name__)
        logger.warning(message)

    @classmethod
    def error(cls, message: str):
        logger = getLogger(cls.__name__)
        logger.error(message)


class ShodanDataHandler(ABC):
    """Abstract base class indicating that a class implements from_shodan()."""

    @classmethod
    def from_shodan(cls, d: Dict):
        pass


class CensysDataHandler(ABC):
    """Abstract base class indicating that a class implements from_censys()."""

    @classmethod
    def from_censys(cls, d: Dict):
        pass


class BinaryEdgeDataHandler(ABC):
    """Abstract base class indicating that a class implements from_binaryedge()."""

    @classmethod
    def from_binaryedge(cls, d: Union[Dict, List]):
        pass



import warnings
from datetime import datetime

from DateTime import DateTime
from mmh3 import hash as mmh3_hash

from common_osint_model.utils import flatten, unflatten, common_model_cn_extraction, sha256_from_body_string, \
    list_cleanup


def from_censys_ipv4(raw: dict) -> dict:
    """
    Converts a Censys IPv4 dictionary into the common format
    :param raw: Censys IPv4 dict
    :return: Common format dict
    """
    warnings.warn("This function was deprecated in v0.4.0. and will be removed in v0.5.0", DeprecationWarning)
    flattened = False
    for k in raw.keys():
        if "." in k:
            flattened = True
            break
        elif k == "443" or k == "80" or k == "22" or k == "autonomous_system":
            break

    if flattened:
        raw = unflatten(raw)

    g = {}
    ports = []
    g.update(censys_ipv4_meta_extraction(raw))
    for protocol in raw.get("protocols", []):
        (port, protocol) = protocol.split("/")
        ports.append(port)
        g.update(censys_ipv4_service_extraction(raw, port, protocol))
    g.update(dict(ports=ports))
    g["domains"] = common_model_cn_extraction(g)
    return g


def from_censys_ipv4_flattened(raw: dict) -> dict:
    """
    Converts a Censys IPv4 dictionary into the common format
    :param raw: Censys IPv4 dict
    :return: Common format dict, flattened
    """
    warnings.warn("This function was deprecated in v0.4.0. and will be removed in v0.5.0", DeprecationWarning)
    return flatten(from_censys_ipv4(raw))


def censys_ipv4_meta_extraction(raw: dict) -> dict:
    """
    Extracts metadata from Censys IPv4 dicts
    :param raw: Censys IPv4 dict
    :return: Metadata part of common format dict
    """
    _as = raw.get("autonomous_system", None) or dict()
    return {
        "ip": raw["ip"],  # Program should fail if IP is not given
        "as": {
            "number": _as.get("asn", None),
            "name": _as.get("name", None),
            "location": _as.get("country_code", None),
            "prefix": _as.get("routed_prefix", None),
        },
    }


def censys_ipv4_service_extraction(raw: dict, port: str, protocol: str) -> dict:
    """
    Extracts meta information and routes to correct method for service detail extraction.
    :param raw: Censys IPv4 dict
    :param port: Port as str
    :param protocol: Protocol as str
    :return: Service dictionary
    """
    s = raw.get(port, {}).get(protocol, None) or dict()
    timestamp = raw.get("updated_at", None)
    service = {
        "timestamp": int(DateTime(timestamp)) if timestamp else None,
        "timestamp_readable": DateTime(timestamp).ISO8601() if timestamp else None,
    }
    keys = s.keys()
    if "banner_decoded" in keys:
        service.update(dict(banner=s["banner_decoded"]))
    if "get" in keys:
        http = censys_ipv4_http_extraction(s["get"])
        service.update({"http": http})
    if "tls" in keys:
        tls = censys_ipv4_tls_extraction(s["tls"])
        service.update({"tls": tls})
    if protocol == 'ssh':
        ssh = censys_ipv4_ssh_extraction(s)
        service.update({"ssh": ssh})
    return {port: service}


def censys_ipv4_http_extraction(s: dict) -> dict:
    """
    Extracts HTTP relevant data out ot service part of Censys IPv4 dict
    :param s: Service part of a censys dict
    :return: Dictionary with HTTP data
    """
    headers = s.get("headers", {})
    for h in headers.get("unknown", []):
        headers.update({h["key"].lower().replace("-", "_"): h["value"]})
    if "unknown" in headers.keys():
        del headers["unknown"]
    return {
        "headers": headers,
        "content": {
            "html": s.get("body", None),
            "hash": {
                "shodan": mmh3_hash(s.get("body", None) or ""),
                "sha256": s.get("body_sha256", None) or sha256_from_body_string("")
            },
            "favicon": {"shodan": None, "sha256": None},
        },
    }


def censys_ipv4_tls_extraction(s: dict) -> dict:
    """
    Extracts TLS relevant data out ot service part of Censys IPv4 dict
    :param s: Service part of a censys dict
    :return: Dictionary with TLS data
    """
    cert = s.get("certificate", {}).get("parsed", {})
    subject = cert.get("subject", None) or dict()
    issuer = cert.get("issuer", None) or dict()
    validity = cert.get("validity", None) or dict()
    common_name = subject.get("common_name", [])
    common_name.extend(cert.get("names", []))
    if len(common_name) == 0:
        common_name = None
    cert_issued = DateTime(validity.get("start", None))
    cert_expires = DateTime(validity.get("end", None))
    cert_length = validity.get("length", None)
    return {
        "certificate": {
            "issuer_dn": cert.get("issuer_dn", None),
            "subject_dn": cert.get("subject_dn", None),
            "issuer": {
                # Censys always uses lists for those kind of attributes
                "common_name": issuer.get("common_name", [None])[0],
                "country": issuer.get("country", [None])[0],
                "locality": issuer.get("locality", [None])[0],
                "province": issuer.get("province", [None])[0],
                "organization": issuer.get("organization", [None])[0],
                "organizational_unit": issuer.get("organizational_unit", [None])[0],
                "email_address": issuer.get("email_address", [None])[0],
            },
            "subject": {
                # Censys always uses lists for those kind of attributes, multiple CNs are okay, though
                "common_name": common_name,
                "country": subject.get("country", [None])[0],
                "locality": subject.get("locality", [None])[0],
                "province": subject.get("province", [None])[0],
                "organization": subject.get("organization", [None])[0],
                "organizational_unit": subject.get("organizational_unit", [None])[0],
                "email_address": subject.get("email_address", [None])[0],
            },
            "validity": {
                "start": int(cert_issued),
                "start_readable": cert_issued.ISO8601(),
                "end": int(cert_expires),
                "end_readable": cert_expires.ISO8601(),
                "length": cert_length
            },
            "fingerprint": {
                "sha1": cert.get("fingerprint_sha1", None),
                "sha256": cert.get("fingerprint_sha256", None)
            }
        }
    }


def censys_ipv4_ssh_extraction(s: dict) -> dict:
    """
    Extracts SSH relevant data out ot service part of Censys IPv4 dict
    :param s: Service part of a censys dict
    :return: Dictionary with SSH data
    """
    v2 = s.get("v2", None) or dict()
    banner = v2.get("banner", None) or dict()
    support = v2.get("support", None) or dict()
    s2c = support.get("server_to_client", None) or dict()
    shk = v2.get("server_host_key", None) or dict()
    return {
        "version": banner.get("raw", None),
        "key_exchange": {
            "algorithms": {
                "compression": s2c.get("compressions", None),
                "encryption": s2c.get("ciphers", None),
                "key_exchange": support.get("kex_algorithms", None),
                "mac": s2c.get("macs", None),
                "key_algorithms": support.get("host_key_algorithms", None)
            }
        },
        "key": {
            "hash": {
                "sha256": shk.get("fingerprint_sha256", None)
            },
            "type": shk.get("key_algorithm", None)
        }
    }


def from_censys_certificates(raw: dict) -> dict:
    """
    Converts a censys certificates dictionary to the common model format
    :param raw: Censys certificates dict
    :return: Common model dict
    """
    flattened = False
    for k in raw.keys():
        if "." in k:
            flattened = True
            break
        elif k == "parsed":
            break

    if flattened:
        raw = unflatten(raw)

    return dict(censys_certificates_parsed_extraction(raw["parsed"]))


def from_censys_certificates_flattened(raw: dict) -> dict:
    """
    Converts a censys certificates dictionary to the flattened common model format
    :param raw: Censys certificates dict
    :return: Common model dict
    """
    return flatten(from_censys_certificates(raw))


def censys_certificates_parsed_extraction(parsed: dict) -> dict:
    """
    Extracts the parsed certificate data
    :param p: "Parsed" dictionary of censys data
    :return: Dictionary with parsed certificate data in common model format
    """
    issuer = {}
    subject = {}
    fingerprint = {}
    validity = {}
    for issuer_key, issuer_item in parsed.get("issuer", {}).items():
        issuer.update({
            issuer_key: issuer_item
        })
    for subject_key, subject_item in parsed.get("subject", {}).items():
        subject.update({
            subject_key: subject_item
        })
    for fp_key, fp_item in parsed.items():
        if "fingerprint_" not in fp_key:
            continue
        fingerprint.update({
            fp_key.replace("fingerprint_", ""): fp_item
        })

    for extension, content in parsed.get("extensions", {}).items():
        if extension == "subject_alt_name":
            cn = subject.get("common_name", None)
            altnames = content.get("dns_names", [])
            if cn and cn in altnames or not cn:
                subject["common_name"] = altnames
            elif cn and cn not in altnames:
                if isinstance(cn, str):
                    altnames.append(cn)
                elif isinstance(cn, list):
                    for name in cn:
                        if name not in altnames:
                            altnames.append(name)
                subject["common_name"] = altnames

    start = datetime.strptime(parsed["validity"]["start"], "%Y-%m-%dT%H:%M:%SZ")
    end = datetime.strptime(parsed["validity"]["end"], "%Y-%m-%dT%H:%M:%SZ")
    validity = {
        "start": int(start.timestamp()),
        "start_readable": start.isoformat(),
        "end": int(end.timestamp()),
        "end_readable": end.isoformat(),
        "length": int((end - start).total_seconds())
    }

    p = {
        "issuer": issuer,
        "issuer_dn": parsed.get("issuer_dn", None),
        "subject": subject,
        "subject_dn": parsed.get("subject_dn", None),
        "fingerprint": fingerprint,
        "validity": validity,
        "serial_number": parsed.get("serial_number", None)
    }
    return list_cleanup(p)



import re
import warnings
from base64 import b64decode
from binascii import hexlify
from datetime import datetime
from hashlib import sha256
from typing import Union

from DateTime import DateTime

from common_osint_model.utils import flatten, common_model_cn_extraction, sha256_from_body_string


def from_shodan(raw: Union[list, dict]) -> dict:
    """
    Turn an expected Shodan dictionary into a generic format
    :param raw: Shodan list or dictionary from host queries
    :return: Generic host describing dictionary
    """
    warnings.warn("This function was deprecated in v0.4.0. and will be removed in v0.5.0", DeprecationWarning)
    g = {}
    services = []
    if isinstance(raw, dict):
        if isinstance(raw.get("data", ""), list):
            services.extend(raw["data"])
        else:
            services.append(raw)
    elif isinstance(raw, list):
        if len(raw) == 1:
            raw = raw[0]
            if isinstance(raw.get("data", ""), list):
                services.extend(raw["data"])
            else:
                services.append(raw)
        else:
            services.extend(raw)
    else:
        raise TypeError("Given parameter 'raw' must be either a list or a dict.")
    # Get the meta data
    g.update(shodan_meta_extraction(services))
    # Get data for every service
    ports = []
    for s in services:
        ports.append(s["port"])
        g.update(shodan_service_extraction(s))
    g.update(dict(ports=ports))
    g["domains"].extend(common_model_cn_extraction(g))
    return g


def from_shodan_flattened(raw: Union[list, dict]) -> dict:
    """
    Turn an expected Shodan dictionary into a generic format
    :param raw: Shodan list or dictionary from host queries
    :return: Generic host describing dictionary, flattened
    """
    warnings.warn("This function was deprecated in v0.4.0. and will be removed in v0.5.0", DeprecationWarning)
    return flatten(from_shodan(raw))


def shodan_meta_extraction(raw: Union[dict, list]) -> dict:
    """
    Returns a dictionary containing all the meta information from a Shodan host or service
    :param raw: Either a service dict or a host list object
    :return: Dictionary containing all meta information
    """
    g = {}
    if isinstance(raw, dict):
        o = raw
    elif isinstance(raw, list):
        o = raw[0]
    else:
        raise TypeError("Given parameter 'raw' must be either a list or a dict.")
    g.update(
        {
            "as": {
                "number": o.get("asn", None),
                "name": o.get("isp", None),
                "location": o.get("location", {}).get("country_code", None),
                "prefix": None,  # Todo: Check of routed prefix is given in censys output
            },
            "domains": [],
            "org": o.get("org", None),
            "ip": o.get("ip_str"),
            "location": {
                "city": o.get("city", None),
                "country": o.get("country_name", None),
                "country_code": o.get("country_code", None),
                "postal_code": o.get("postal_code", None),
                "coordinates": {
                    "latitude": o.get("latitude", None),
                    "longitude": o.get("longitude", None)
                }
            }
        }
    )
    for domain in o.get("domains", []):
        g["domains"].append({"type": "shodan-domain", "value": domain})
    for hostname in o.get("hostnames", []):
        g["domains"].append({"type": "shodan-hostname", "value": hostname})
    return g


def shodan_service_extraction(s: dict) -> dict:
    """
    Returns a dictionary containing a properly formatted service based on a shodan raw dictionary
    :param s: Shodan service dictionary
    :return: Porperly formatted dictionary
    """
    g = {}
    p = s["port"]
    g.update(
        {
            p: {
                "banner": s["data"],
                "timestamp": int(DateTime(s["timestamp"])),
                "timestamp_readable": DateTime(s["timestamp"]).ISO8601(),
            }
        }
    )
    k = s.keys()

    if "http" in k or s["data"][:4] == "HTTP":
        g[p].update(shodan_http_extraction(s))
    if "ssl" in k:
        g[p].update(shodan_ssl_extraction(s))
    if "ssh" in k:
        g[p].update(shodan_ssh_extraction(s))
    return g


def shodan_http_extraction(s: dict) -> dict:
    """
    Extracts http(s) specific values from shodan service dict
    :param s: Shodan service dictionary
    :return: Properly formatted, service related dictionary
    """
    http = s.get("http", None) or dict()
    headers = {}
    for line in s["data"].split("\n"):
        line = line.strip()
        if not ":" in line:
            continue
        h = line.split(":")
        headers.update({h[0].lower().replace("-", "_"): ":".join(h[1:]).strip()})
    favicon = http.get("favicon", None) or dict()
    favicon_data = favicon.get("data", None)
    favicon_hash = favicon.get("hash", None)
    favicon_sha256 = None
    if favicon:
        favicon_decoded = b64decode(favicon_data)
        h = sha256()
        h.update(favicon_decoded)
        favicon_sha256 = hexlify(h.digest()).decode("ascii")
    return {
        "http": {
            "headers": headers,
            "content": {
                "html": http.get("html", None),
                "hash": {
                    "shodan": http.get("html_hash", None),
                    "sha256": sha256_from_body_string(http.get("html", None) or "")
                },
                "favicon": {"shodan": favicon_hash, "sha256": favicon_sha256},
            },
        }
    }


def shodan_ssl_extraction(s: dict) -> dict:
    """
    Extracts ssl/tls specific values from shodan service dict
    :param s: Shodan service dictionary
    :return: Properly formatted, service related dictionary
    """
    ssl = s.get("ssl", None) or dict()
    cert = ssl.get("cert", {})
    subject = cert.get("subject", None) or dict()
    issuer = cert.get("issuer", None) or dict()
    fingerprint = cert.get("fingerprint", None) or dict()
    dhparams = ssl.get("dhparams", None) or dict()
    common_name = subject.get("CN", None)
    common_names = set()
    if common_name:
        common_names.add(common_name)
    for ext in ssl.get("extensions", []):
        if ext["name"] == "subjectAltName":
            for domain in re.split(r"(\\x[0-9a-z]{2}|\\[a-z^x]{1})", ext["date"]):
                if "." in domain:
                    common_names.add(domain)
    cert_issued = int(datetime.strptime(cert.get("issued", None), "%Y%m%d%H%M%SZ").timestamp())
    cert_expires = int(datetime.strptime(cert.get("expires", None), "%Y%m%d%H%M%SZ").timestamp())
    return {
        "tls": {
            "certificate": {
                "subject": {
                    "country": subject.get("C", None),
                    "province": subject.get("ST", None),
                    "locality": subject.get("L", None),
                    "common_name": list(common_names),
                    "organization": subject.get("O", None),
                    "organizational_unit": subject.get("OU", None),
                    "email_address": subject.get("email", None),
                },
                "issuer": {
                    "country": issuer.get("C", None),
                    "province": issuer.get("ST", None),
                    "locality": issuer.get("L", None),
                    "common_name": issuer.get("CN", None),
                    "organization": issuer.get("O", None),
                    "organizational_unit": issuer.get("OU", None),
                    "email_address": issuer.get("email", None),
                },
                "fingerprint": {
                    "sha1": fingerprint.get("sha1", None),
                    "sha256": fingerprint.get("sha256", None),
                },
                "serial_number": str(int(cert.get("serial", None))),
                "validity": {
                    "start": cert_issued,
                    "start_readable": DateTime(cert_issued, 'UTC').ISO8601(),
                    "end": cert_expires,
                    "end_readable": DateTime(cert_expires, 'UTC').ISO8601(),
                    "length": cert_expires - cert_issued,
                }
            },
            "dhparam": {
                "bits": dhparams.get("bits", None),
                "generator": dhparams.get("generator", None),
            },
            "cipher": {
                "id": None,
                "name": ssl.get("cipher", {}).get("name", None),
                "bits": ssl.get("cipher", {}).get("bits", None),
            },
            "ja3s": ssl.get("ja3s", None),
            "jarm": ssl.get("jarm", None)
        }
    }


def shodan_ssh_extraction(s: dict) -> dict:
    """
    Extracts ssh specific values from shodan service dict
    :param s: Shodan service dictionary
    :return: Properly formatted, service related dictionary
    """
    ssh = s.get("ssh", None) or dict()
    key_exchange = ssh.get("kex", None) or dict()
    h = sha256()
    h.update(b64decode(ssh.get("key", None)))
    return {
        "ssh": {
            "version": s["data"].split("Key type")[0].strip(),
            "key_exchange": {
                "algorithms": {
                    "compression": key_exchange.get("compression_algorithms", None),
                    "encryption": key_exchange.get("encryption_algorithms", None),
                    "key_exchange": key_exchange.get("kex_algorithms", None),
                    "mac": key_exchange.get("mac_algorithms", None),
                    "key_algorithms": key_exchange.get(
                        "server_host_key_algorithms", None
                    ),
                }
            },
            "key": {
                "hash": {
                    "sha256": hexlify(h.digest()).decode("ascii")
                },
                "type": ssh.get("type", None),
            },
        }
    }



import base64
from typing import Dict, List, Optional, Union

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicNumbers
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from pydantic import BaseModel

from common_osint_model.models import ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger
from common_osint_model.utils import hash_all


class SSHComponentAlgorithms(BaseModel, ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger):
    """Represents algorithms supported by SSH server."""
    encryption: Optional[List[str]]
    key_exchange: Optional[List[str]]
    mac: Optional[List[str]]
    key_algorithms: Optional[List[str]]
    compression: Optional[List[str]]

    @classmethod
    def from_shodan(cls, d: Dict) -> Union["SSHComponentAlgorithms", None]:
        """Returns an instance of this class based on Shodan data given as dictionary."""
        if not isinstance(d, Dict):
            raise TypeError(f"Method SSHComponentAlgorithms.from_shodan expects parameter d to be a dictionary, "
                            f"but it was {type(d)}.")
        try:
            return SSHComponentAlgorithms(
                encryption=d["ssh"]["kex"]["encryption_algorithms"],
                key_exchange=d["ssh"]["kex"]["kex_algorithms"],
                mac=d["ssh"]["kex"]["mac_algorithms"],
                key_algorithms=d["ssh"]["kex"]["server_host_key_algorithms"],
                compression=d["ssh"]["kex"]["compression_algorithms"]
            )
        except KeyError as ke:
            cls.warning(f"Shodan data is missing key: {ke}")
            return None

    @classmethod
    def from_censys(cls, d: Dict) -> Union["SSHComponentAlgorithms", None]:
        """Returns an instance of this class based on Censys data given as dictionary."""
        try:
            return SSHComponentAlgorithms(
                encyption=d["ssh"]["kex_init_message"]["client_to_server_ciphers"],
                key_exchange=d["ssh"]["kex_init_message"]["kex_algorithms"],
                mac=d["ssh"]["kex_init_message"]["server_to_client_macs"],
                key_algorithms=d["ssh"]["kex_init_message"]["host_key_algorithms"],
                compression=d["ssh"]["kex_init_message"]["server_to_client_compression"]
            )
        except KeyError as ke:
            cls.warning(f"Censys data is missing key: {ke}")
            return None

    @classmethod
    def from_binaryedge(cls, d: Dict) -> Union["SSHComponentAlgorithms", None]:
        """Returns an instance of this class based on BinaryEdge data given as dictionary."""
        try:
            return SSHComponentAlgorithms(
                encryption=d["encryption"],
                key_exchange=d["kex"],
                mac=d["mac"],
                key_algorithms=d["server_host_key"],
                compression=d["compression"]
            )
        except KeyError as ke:
            cls.warning(f"BinaryEdge data is missing key: {ke}")
            return None


class SSHComponentKey(BaseModel, ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger):
    """Represents the public key exposed by the SSH server."""
    # Type represents the ssh-key type, e.g. ssh-rsa
    raw: Optional[str]
    type: Optional[str]
    md5: Optional[str]
    sha1: Optional[str]
    sha256: Optional[str]
    murmur: Optional[str]

    @classmethod
    def from_shodan(cls, d: Dict) -> Union["SSHComponentKey", None]:
        """Returns an instance of this class based on Shodan data given as dictionary."""
        if not isinstance(d, Dict):
            raise TypeError(f"Method SSHComponentKey.from_shodan expects parameter d to be a dictionary, but it was "
                            f"{type(d)}.")

        try:
            key = d["ssh"]["key"]
            key = base64.b64decode(key)
            md5, sha1, sha256, murmur = hash_all(key)
            return SSHComponentKey(
                raw=d["ssh"]["key"],
                type=d["ssh"]["type"],
                md5=md5,
                sha1=sha1,
                sha256=sha256,
                murmur=murmur
            )
        except KeyError as ke:
            cls.warning(f"Shodan data is missing key: {ke}")
            return None

    @classmethod
    def from_censys(cls, d: Dict) -> Union["SSHComponentKey", None]:
        """Returns an instance of this class based on Censys data given as dictionary."""
        cls.info("Censys data does not contain the key as raw data. The public key can be constructed with given "
                 "data, however, currently this is only supported for RSA keys.")

        try:
            if "rsa_public_key" in d["ssh"]["server_host_key"]:
                cls.debug("Seems to be a RSA key. Trying to create public key from modulus and exponent.")
                public_numbers = RSAPublicNumbers(
                    e=int.from_bytes(
                        base64.b64decode(d["ssh"]["server_host_key"]["rsa_public_key"]["exponent"]),
                        byteorder="big",
                        signed=False
                    ),
                    n=int.from_bytes(
                        base64.b64decode(d["ssh"]["server_host_key"]["rsa_public_key"]["modulus"]),
                        byteorder="big",
                        signed=False
                    )
                )
                public_key = public_numbers.public_key()
                public_key_string = public_key.public_bytes(
                    encoding=Encoding.OpenSSH,
                    format=PublicFormat.OpenSSH
                ).decode("utf-8")
                cls.debug(f"Created public key from modulus and exponent: {public_key_string}")
                public_key_b64 = public_key_string.split(" ", maxsplit=1)[1]
                public_key_raw_data = base64.b64decode(public_key_b64)
                md5, sha1, sha256, murmur = hash_all(public_key_raw_data)
                return SSHComponentKey(
                    raw=public_key_b64,
                    type="ssh-rsa",
                    md5=md5,
                    sha1=sha1,
                    sha256=sha256,
                    murmur=murmur
                )
            else:
                key_type = "unknown"
                for key in d["ssh"]["server_host_key"].keys():
                    if "public_key" in key:
                        key_type = key.replace("_public_key", "")

                cls.info(f"SSH key type is {key_type}. Currently, only RSA SSH keys are supported in Censys model.")
                return SSHComponentKey(
                    type=key_type,
                    sha256=d["ssh"]["server_host_key"]["fingerprint_sha256"]
                )
        except KeyError as ke:
            cls.warning(f"Censys data is missing key: {ke}")
            return None

    @classmethod
    def from_binaryedge(cls, d: Dict) -> Union["SSHComponentKey", None]:
        """Returns an instance of this class based on BinaryEdge data given as dictionary."""
        try:
            public_key_raw_data = base64.b64decode(d["key"])
            md5, sha1, sha256, murmur = hash_all(public_key_raw_data)
            return SSHComponentKey(
                raw=d["key"],
                type=d["cypher"],
                md5=md5,
                sha1=sha1,
                sha256=sha256,
                murmur=murmur
            )
        except KeyError as ke:
            cls.warning(f"BinaryEdge data is missing key: {ke}")
            return None


class SSHComponent(BaseModel, ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger):
    """Represents the SSH component of services."""
    algorithms: Optional[SSHComponentAlgorithms]
    key: Optional[SSHComponentKey]
    hassh: Optional[str]

    @classmethod
    def from_shodan(cls, d: Dict):
        """Creates an instance of this class based on Shodan data given as dictionary."""
        if not isinstance(d, Dict):
            raise TypeError(f"Method SSHComponent.from_shodan expects parameter d to be a dictionary, "
                            f"but it was {type(d)}.")

        hassh = d.get("ssh", {}).get("hashh", None)

        return SSHComponent(
            algorithms=SSHComponentAlgorithms.from_shodan(d),
            key=SSHComponentKey.from_shodan(d),
            hassh=hassh
        )

    @classmethod
    def from_censys(cls, d: Dict):
        """Creates an instance of this class based on Censys data given as dictionary."""
        hassh = d.get("ssh", {}).get("hassh_fingerprint", None)
        return SSHComponent(
            algorithms=SSHComponentAlgorithms.from_censys(d),
            key=SSHComponentKey.from_censys(d),
            hassh=hassh
        )

    @classmethod
    def from_binaryedge(cls, d: Dict) -> Union["SSHComponent", None]:
        """Creates an instance of this class based on BinaryEdge data given as dictionary."""
        try:
            cyphers = d["result"]["data"]["cyphers"]
            algorithms = d["result"]["data"]["algorithms"]
            hassh = d["result"]["data"]["hassh"]["hassh"]
            cypher = None
            for c in cyphers:
                if c["cypher"] == "ssh-dss":
                    continue
                cypher = c
            return SSHComponent(
                algorithms=SSHComponentAlgorithms.from_binaryedge(algorithms),
                key=SSHComponentKey.from_binaryedge(cypher),
                hassh=hassh
            )
        except KeyError as ke:
            cls.warning(f"BinaryEdge data is missing key: {ke}")
            return None



from pydantic import BaseModel
from typing import Optional, Dict
from common_osint_model.models import ShodanDataHandler, CensysDataHandler


class DNSComponent(BaseModel, ShodanDataHandler, CensysDataHandler):
    recursive: Optional[bool]

    @classmethod
    def from_shodan(cls, d: Dict):
        return DNSComponent(
            recursive=d.get("dns", {}).get("recursive", False)
        )

    @classmethod
    def from_censys(cls, d: Dict):
        return DNSComponent(
            recursive=d.get("dns", {}).get("server_type", "") == "FORWARDING"
        )



import ipaddress
import json
from datetime import datetime
from typing import Optional, Dict, List, Union

from pydantic import BaseModel, validator

from common_osint_model.models import ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger
from common_osint_model.models.autonomous_system import AutonomousSystem
from common_osint_model.models.domain import Domain
from common_osint_model.models.service import Service
from common_osint_model.utils import flatten


class Host(BaseModel, ShodanDataHandler, CensysDataHandler, BinaryEdgeDataHandler, Logger):
    """This class represents a host and can be used to handle results from the common model in a pythonic way."""
    ip: str
    # Information about the autonomous system the host is assigned to
    autonomous_system: Optional[AutonomousSystem]
    # List of services running (listening) on the IP
    services: Optional[List[Service]]
    # List of open ports also mentioned in the open
    ports: Optional[List[int]]
    # Timestamps for activity tracking
    first_seen: Optional[datetime] = datetime.utcnow()
    last_seen: Optional[datetime] = datetime.utcnow()
    # A list of domains, fqdns, common names - or other attributes which represent domainnames -  assigned to the host
    domains: Optional[List[Domain]]
    # This represents the source where the host information was obtained, e.g. shodan, censys...
    source: Optional[str]
    # Optionally, the used query to find the host can be assigned to the object also which might be useful for comparing
    # different hosts later on
    query: Optional[str]

    @validator("ip")
    def validates_ip(cls, v):
        try:
            ipaddress.ip_address(v)
        except Exception as e:
            raise ValueError(f"\"{v}\" is not a correct IP address or at least it is not parseable with the ipaddress"
                             f"module: {e}")
        return v

    @property
    def services_dict(self):
        """Returns the services as dictionary in the form of {port: service}. Uses exclude_none to skip empty keys."""
        # Load the JSON dump, so datetime objects are in iso format.
        json_dict = json.loads(self.json(exclude_none=True))
        json_dict.update({s["port"]: s for s in json_dict["services"]})
        del json_dict["services"]
        return json_dict

    @property
    def flattened_dict(self):
        """Dict in the flattened format."""
        return flatten(self.services_dict)

    @property
    def service_ports(self):
        """Dynamic attribute which loops over available services and grabs the port number. This is kind of redundant
        to the ports attribute, if given, but can help to easily get the values needed for the attribute. Unfortunately
        Pydantic does not support these kind of properties in the data model right now."""
        return [service.port for service in self.services]

    def flattened_json(self) -> str:
        """Returns in the structure formally introduced with the common model."""
        return json.dumps(self.flattened_dict, indent=2)

    @classmethod
    def from_shodan(cls, d: Dict, skip_shodan_domains: Optional[bool] = False):
        if "data" in d and isinstance(d["data"], List):
            d = d["data"]
        domains = []
        domain_strings = []
        if isinstance(d, List):
            for entry in d:
                if "domains" in entry and not skip_shodan_domains:
                    for domain in entry["domains"]:
                        if domain not in domain_strings:
                            domain_strings.append(domain)
                            domains.append(Domain(domain=domain, source="shodan", type="domain"))
                # Check Shodans reverse dns lookups
                if "hostnames" in entry and not skip_shodan_domains:
                    for hostname in entry["hostnames"]:
                        if hostname not in domain_strings:
                            domain_strings.append(hostname)
                            domains.append(Domain(domain=hostname, source="shodan", type="rdns"))
            ip = d[0]["ip_str"]
            services = [Service.from_shodan(service) for service in d]
        else:
            ip = d["ip_str"]
            services = [Service.from_shodan(d)]
        for service in services:
            if service.tls:
                for domain in service.tls.certificate.domains:
                    if domain not in domain_strings:
                        domain_strings.append(domain)
                        domains.append(Domain(
                            domain=domain,
                            first_seen=service.tls.certificate.issued,
                            last_seen=service.tls.certificate.expires,
                            source="shodan",
                            type="common_name"
                        ))
        autonomous_system = AutonomousSystem.from_shodan(d)
        return Host(
            ip=ip,
            autonomous_system=autonomous_system,
            services=services,
            domains=domains,
            source="shodan",
            ports=[service.port for service in services]
        )

    @classmethod
    def from_censys(cls, d: Dict):
        ip = d["ip"]
        services = []
        for service in d["services"]:
            services.append(Service.from_censys(service))

        domains = []
        domain_strings = []
        for service in services:
            if service.tls:
                for domain in service.tls.certificate.domains:
                    if domain not in domain_strings:
                        domain_strings.append(domain)
                        domains.append(Domain(
                            domain=domain,
                            # Currently not given by API
                            # first_seen=service.tls.certificate.issued,
                            # last_seen=service.tls.certificate.expires,
                            source="censys",
                            type="common_name"
                        ))
        return Host(
            ip=ip,
            autonomous_system=AutonomousSystem.from_censys(d),
            services=services,
            domains=domains,
            source="censys",
            ports=[service.port for service in services]
        )

    @classmethod
    def from_binaryedge(cls, d: Union[Dict, List]):
        """This can either be a complete query result, or a list of services running on the same ip."""
        if isinstance(d, Dict) and "results" in d:
            # This is a complete result dictionary, extract the list of services.
            d = d["results"][list(d["results"].keys())[0]]
        elif isinstance(d, Dict) and "events" in d:
            d = d["events"]

        services = {}
        for service in d:
            port = service["target"]["port"]
            if port not in services:
                services[port] = [service]
            else:
                services[port].append(service)
        services_objects = [Service.from_binaryedge(service) for service in services.values()]
        ip = d[0]["target"]["ip"]
        domains = []
        domain_strings = []
        for service in services_objects:
            if service.tls:
                for domain in service.tls.certificate.domains:
                    if domain not in domain_strings:
                        domain_strings.append(domain)
                        domains.append(Domain(
                            domain=domain,
                            first_seen=service.tls.certificate.issued,
                            last_seen=service.tls.certificate.expires,
                            source="binaryedge",
                            type="common_name"
                        ))
        return Host(
            ip=ip,
            services=services_objects,
            domains=domains,
            source="binaryedge",
            ports=[service.port for service in services_objects]
        )



from cryptography.x509 import load_pem_x509_certificate
from cryptography.x509.extensions import SubjectAlternativeName
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.hashes import MD5, SHA1, SHA256
from cryptography.hazmat.backends.openssl.x509 import _Certificate as Certificate
from binascii import hexlify
from typing import Union
from datetime import timezone
from common_osint_model.utils import flatten, list_cleanup


def from_x509_pem(pem: Union[str, bytes]) -> dict:
    """
    Converts a certificate in PEM format given as bytes or as string to a dictionary.
    :param pem: PEM certificates as either string or bytes
    :return: dictionary in common format
    """
    g = {}
    if isinstance(pem, str):
        pem = pem.encode("ascii")

    certificate = load_pem_x509_certificate(pem, default_backend())
    g.update(certificate_dn_extraction(certificate))
    g.update(certificate_fingerprint_extraction(certificate))
    g.update(certificate_validity_extraction(certificate))
    g.update(dict(serial_number=certificate.serial_number))
    return list_cleanup(g)


def from_x509_pem_flattened(pem: Union[str, bytes]) -> dict:
    """
    Wraps from_x509_pem and flattens the output dict
    :param pem: PEM certificates as either string or bytes
    :return: flattened dictionary in common format
    """
    return flatten(from_x509_pem(pem))


def certificate_dn_extraction(certificate: Certificate) -> dict:
    """
    Extracts distinguished names of the given certificate
    :param certificate: object of type cryptography.hazmat.backends.openssl.x509._Certificate
    :return: dictionary containing issuer and subject DN
    """
    dns = dict(subject={}, issuer={}, subject_dn=None, issuer_dn=None)
    terms = dict(
        CN='common_name',
        C='country',
        L='locality',
        ST='province',
        O='organization',
        OU='organizational_unit',
        email='email_address'
    )

    dns["issuer_dn"] = certificate.issuer.rfc4514_string()
    for term in certificate.issuer.rfc4514_string().split(','):
        k, v = term.split("=")
        key = terms[k.strip()]
        if key in dns["issuer"].keys():
            if isinstance(dns["issuer"][key], list):
                dns["issuer"][key].append(v)
            else:
                dns["issuer"][key] = [dns["issuer"][key], v]
        else:
            dns["issuer"].update({
                terms[k.strip()]: v
            })

    dns["subject_dn"] = certificate.subject.rfc4514_string()
    for term in certificate.subject.rfc4514_string().split(','):
        k, v = term.split("=")
        key = terms[k.strip()]
        if key in dns["subject"].keys():
            if isinstance(dns["subject"][key], list):
                dns["subject"][key].append(v)
            else:
                dns["subject"][key] = [dns["subject"][key], v]
        else:
            dns["subject"].update({
                terms[k.strip()]: v
            })
    try:
        subjectAltName = certificate.extensions.get_extension_for_oid(SubjectAlternativeName.oid)
    except:
        subjectAltName = None

    if subjectAltName:
        dns["subject"]["common_name"] = [dns["subject"]["common_name"]]
        for v in subjectAltName.value:
            if v.value not in dns["subject"]["common_name"]:
                dns["subject"]["common_name"].append(v.value)

    return dns


def certificate_fingerprint_extraction(certificate: Certificate) -> dict:
    """
    Calculates certificate fingerprints as MD5, SHA1 and SHA256
    :param certificate: object of type cryptography.hazmat.backends.openssl.x509._Certificate
    :return: dictionary containing all fingerprints
    """
    return {
        "fingerprint": {
            "md5": hexlify(certificate.fingerprint(MD5())).decode("ascii"),
            "sha1": hexlify(certificate.fingerprint(SHA1())).decode("ascii"),
            "sha256": hexlify(certificate.fingerprint(SHA256())).decode("ascii"),
        }
    }


def certificate_validity_extraction(certificate: Certificate) -> dict:
    """
    Extracts validity information of given certificate
    :param certificate: object of type cryptography.hazmat.backends.openssl.x509._Certificate
    :return: dictionary containing the validity timestamps
    """
    return {
        "validity": {
            "start": int(certificate.not_valid_before.timestamp()),
            "start_readable": certificate.not_valid_before.replace(tzinfo=timezone.utc, microsecond=0).isoformat(),
            "end": int(certificate.not_valid_after.timestamp()),
            "end_readable": certificate.not_valid_after.replace(tzinfo=timezone.utc, microsecond=0).isoformat(),
            "length": int((certificate.not_valid_after - certificate.not_valid_before).total_seconds())
        }
    }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['common_osint_model', 'common_osint_model.censys', 'common_osint_model.models']

package_data = \
{'': ['*']}

install_requires = \
['DateTime>=4.3,<5.0',
 'cryptography>=38.0.3,<39.0.0',
 'mmh3>=3.0.0,<4.0.0',
 'pydantic>=1.8.2,<2.0.0']

entry_points = \
{'console_scripts': ['convcensys = common_osint_model.cli:convcensys',
                     'convcensys2 = common_osint_model.cli:convcensys2',
                     'convcensyscert = common_osint_model.cli:convcensyscert',
                     'convcert = common_osint_model.cli:convcert',
                     'convshodan = common_osint_model.cli:convshodan']}

setup_kwargs = {
    'name': 'common-osint-model',
    'version': '0.4.13',
    'description': 'Converting data from services like BinaryEdge, Censys and Shodan to a common data model.',
    'long_description': '# Common OSINT Model\n\n**Note:** This is work in progress and probably only covers my specific use case. If you find bugs or know how to\nenhance this project, please open an issue or - even better - create a pull request. The presented data model is\nevolving continuously. Therefore, it is recommended to use it in your project with a fixed version constraint (e.g. \n`common-osint-model==0.4.1`) and take a look at what has changed here before updating `common-osint-model` as a\ndependency.\n\nThis project aims to create an easy to use data model as well as implement converters for commonly used sources. As my\nuse case often includes HTTP(S), TLS and SSH only, data delivered for other protocols by the given sources might not\nshow up correctly or just by the banner included. Because the available scanning services use different hashes for \ndifferent purposes, one might be not able to search across services with the available data. This model automatically \ncalculates hashes wherever raw data is available.\n\n## Todos\n- [ ] Implement additional SSH ciphers for Censys data. Currently only RSA is supported. Censys do not provide raw \n public keys in their data, but the public key can be built through the components given.\n\n## The data model\n\n**Note:** The model was restructured with version 0.4.0. The previously provieded functionality can still be used,\nhowever, the old model is considered deprecated and is not further developed.\n\nSince version 0.4.0 the model changed. During refactoring, all model components were implemented as\n[pydantic](https://pydantic-docs.helpmanual.io/) classes in order to provide a more "pythonic" way of interacting with\nthe data. Also, this enables an easy to read and structurized view of the model itself. Please take a look at the model\nfiles available in [models](common_osint_model/models) for more details. The model starts with a host which has an IP,\ndomains pointing to it and provides multiple services:\n\n```python\nclass Host:\n    ip: str\n    autonomous_system: Optional[AutonomousSystem]\n    services: List[Service]\n    first_seen: Optional[datetime] = datetime.utcnow()\n    last_seen: Optional[datetime] = datetime.utcnow()\n    domains: Optional[List[Domain]]\n```\n\nA service listens on a specific port, has various components (currently only HTTP, TLS and SSH implemented), a banner\nand a source (e.g. Shodan) as well as some timestamps.\n\n```python\nclass Service:\n    port: int\n    banner: Optional[str]\n    md5: Optional[str]\n    sha1: Optional[str]\n    sha256: Optional[str]\n    murmur: Optional[str]\n    first_seen: Optional[datetime] = datetime.utcnow()\n    last_seen: Optional[datetime] = datetime.utcnow()\n    timestamp: Optional[datetime]\n    http: Optional[HTTPComponent]\n    tls: Optional[TLSComponent]\n    ssh: Optional[SSHComponent]\n    source: str\n```\n\nThe further model classes are designed in a similar way, please have a look at the\n[source files](common_osint_model/models).\n\n## How to use\n\n### Installation\n\n```bash\npip install common-osint-model\n```\n\n### Convert all the things\n\n```python\n# Post-v0.4.0 (Pydantic model)\nimport shodan\nfrom common_osint_model import Host\n\nshodan_client = shodan.Shodan("My API key")\nraw_shodan_response = shodan_client.host("140.82.121.4")\nhost = Host.from_shodan(raw_shodan_response)\n# Similarly:\n# Host.from_censys(raw_censys_response)\n# Host.from_binaryedge(raw_binaryedge_response)\n# Make sure to only pass results for *one* host. Currently there is no functionality provided to wrap around different \n# return types, such as lists of hosts from a query. You need to loop through them yourself.\nprint(f"Got {host.ip}.")\nprint(f"Providing {len(host.services)} services.")\n\nfor idx, service in enumerate(host.services):\n    print(f"Banner for Service {idx + 1}:")\n    print(f"\\t{service.banner}")\n    print(f"\\tMD5: {service.md5}")\n    print(f"\\tSHA1: {service.sha1}")\n    print(f"\\tSHA256: {service.sha256}")\n    print(f"\\tMurmur: {service.murmur}")\n\nprint("Exporting data as flattened json blob for further use, e.g. Elasticsearch indexing...")\nprint(host.flattened_json())\n\n# Pre-v0.4.0 (Dictionary based approach - DEPRECATED)\nfrom common_osint_model import from_shodan, from_shodan_flattened, from_censys_ipv4, from_censys_ipv4_flattened,\n    from_x509_pem, from_x509_pem_flattened\n\nraw_shodan = get_my_shodan_data()\nconverted_s = from_shodan(raw_shodan)  # Deprecated\nflattened_s = from_shodan_flattened(raw_shodan)  # Deprecated\n\nraw_censys_ipv4 = get_my_censys_ipv4_data()\nconverted_c = from_censys_ipv4(raw_censys_ipv4)\nflattened_c = from_censys_ipv4_flattened(raw_censys_ipv4)\n\nraw_certificate = open(\'certificate.pem\').read()\nconverted_certificate = from_x509_pem(raw_shodan)\nflattened_certificate = from_x509_pem_flattened(raw_shodan)\n```\n\n### Example output\n\n```python\nimport shodan\nfrom common_osint_model import Host\n\nshodan_client = shodan.Shodan("My API key")\nraw_shodan_response = shodan_client.host("140.82.121.4")\nhost = Host.from_shodan(raw_shodan_response)\nprint(f"Host: {host.ip}")\nprint(f"AS: {host.autonomous_system.dict(exclude_none=True)}")\n\nfor service in host.services:\n    print(f"Service: {service}")\n\nprint("--- flattened JSON dump ---")\nprint(host.flattened_json())\n```\n\n```\nHost: 140.82.121.4\nAS: {\'number\': 36459, \'name\': \'GitHub, Inc.\', \'country\': \'DE\', \'source\': \'shodan\'}\nService: port=443 banner=\'HTTP/1.1 301 Moved Permanently\\r\\nContent-Length: 0\\r\\nLocation: https://github.com/\\r\\n\\r\\n\' md5=\'d402a6212741f3690b4fa1e46d9bd8b6\' sha1=\'a24eb4ba0332776d38050a7b41d0366742dbf262\' sha256=\'5fbe0315395986d131e4948888987319e88e3e1da6c5460e8a0bf8b7a1e639f0\' murmur=\'-1655207803\' first_seen=datetime.datetime(2021, 8, 20, 16, 10, 27, 656097) last_seen=datetime.datetime(2021, 8, 20, 16, 10, 27, 656099) timestamp=datetime.datetime(2021, 8, 16, 20, 42, 40, 325940) http=HTTPComponent(headers={\'Content-Length\': \'0\', \'Location\': \'https://github.com/\'}, content=HTTPComponentContent(raw=\'\', length=0, md5=\'d41d8cd98f00b204e9800998ecf8427e\', sha1=\'da39a3ee5e6b4b0d3255bfef95601890afd80709\', sha256=\'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855\', murmur=\'0\', favicon=None, robots_txt=None, security_txt=None)) tls=TLSComponent(certificate=TLSComponentCertificate(issuer=TLSComponentCertificateEntity(dn=\'C=US, O=DigiCert, Inc., CN=DigiCert High Assurance TLS Hybrid ECC SHA256 2020 CA1\', country=\'US\', state=None, locality=None, organization=\'DigiCert, Inc.\', organizational_unit=None, common_name=\'DigiCert High Assurance TLS Hybrid ECC SHA256 2020 CA1\', email_address=None), subject=TLSComponentCertificateEntity(dn=\'C=US, ST=California, L=San Francisco, O=GitHub, Inc., CN=github.com\', country=\'US\', state=\'California\', locality=\'San Francisco\', organization=\'GitHub, Inc.\', organizational_unit=None, common_name=\'github.com\', email_address=None), issued=datetime.datetime(2021, 3, 25, 0, 0), expires=datetime.datetime(2022, 3, 30, 23, 59, 59), expired=False, alternative_names=[\'www.github.com\', \'github.com\'], pem=\'-----BEGIN CERTIFICATE-----\\nMIIFBjCCBK2gAwIBAgIQDovzdw2S0Zbwu2H5PEFmvjAKBggqhkjOPQQDAjBnMQsw\\nCQYDVQQGEwJVUzEXMBUGA1UEChMORGlnaUNlcnQsIEluYy4xPzA9BgNVBAMTNkRp\\nZ2lDZXJ0IEhpZ2ggQXNzdXJhbmNlIFRMUyBIeWJyaWQgRUNDIFNIQTI1NiAyMDIw\\nIENBMTAeFw0yMTAzMjUwMDAwMDBaFw0yMjAzMzAyMzU5NTlaMGYxCzAJBgNVBAYT\\nAlVTMRMwEQYDVQQIEwpDYWxpZm9ybmlhMRYwFAYDVQQHEw1TYW4gRnJhbmNpc2Nv\\nMRUwEwYDVQQKEwxHaXRIdWIsIEluYy4xEzARBgNVBAMTCmdpdGh1Yi5jb20wWTAT\\nBgcqhkjOPQIBBggqhkjOPQMBBwNCAASt9vd1sdNJVApdEHG93CUGSyIcoiNOn6H+\\nudCMvTm8DCPHz5GmkFrYRasDE77BI3q5xMidR/aW4Ll2a1A2ZvcNo4IDOjCCAzYw\\nHwYDVR0jBBgwFoAUUGGmoNI1xBEqII0fD6xC8M0pz0swHQYDVR0OBBYEFCexfp+7\\nJplQ2PPDU1v+MRawux5yMCUGA1UdEQQeMByCCmdpdGh1Yi5jb22CDnd3dy5naXRo\\ndWIuY29tMA4GA1UdDwEB/wQEAwIHgDAdBgNVHSUEFjAUBggrBgEFBQcDAQYIKwYB\\nBQUHAwIwgbEGA1UdHwSBqTCBpjBRoE+gTYZLaHR0cDovL2NybDMuZGlnaWNlcnQu\\nY29tL0RpZ2lDZXJ0SGlnaEFzc3VyYW5jZVRMU0h5YnJpZEVDQ1NIQTI1NjIwMjBD\\nQTEuY3JsMFGgT6BNhktodHRwOi8vY3JsNC5kaWdpY2VydC5jb20vRGlnaUNlcnRI\\naWdoQXNzdXJhbmNlVExTSHlicmlkRUNDU0hBMjU2MjAyMENBMS5jcmwwPgYDVR0g\\nBDcwNTAzBgZngQwBAgIwKTAnBggrBgEFBQcCARYbaHR0cDovL3d3dy5kaWdpY2Vy\\ndC5jb20vQ1BTMIGSBggrBgEFBQcBAQSBhTCBgjAkBggrBgEFBQcwAYYYaHR0cDov\\nL29jc3AuZGlnaWNlcnQuY29tMFoGCCsGAQUFBzAChk5odHRwOi8vY2FjZXJ0cy5k\\naWdpY2VydC5jb20vRGlnaUNlcnRIaWdoQXNzdXJhbmNlVExTSHlicmlkRUNDU0hB\\nMjU2MjAyMENBMS5jcnQwDAYDVR0TAQH/BAIwADCCAQUGCisGAQQB1nkCBAIEgfYE\\ngfMA8QB2ACl5vvCeOTkh8FZzn2Old+W+V32cYAr4+U1dJlwlXceEAAABeGq/vRoA\\nAAQDAEcwRQIhAJ7miER//DRFnDJNn6uUhgau3WMt4vVfY5dGigulOdjXAiBIVCfR\\nxjK1v4F31+sVaKzyyO7JAa0fzDQM7skQckSYWQB3ACJFRQdZVSRWlj+hL/H3bYbg\\nIyZjrcBLf13Gg1xu4g8CAAABeGq/vTkAAAQDAEgwRgIhAJgAEkoJQRivBlwo7x67\\n3oVsf1ip096WshZqmRCuL/JpAiEA3cX4rb3waLDLq4C48NSoUmcw56PwO/m2uwnQ\\nprb+yh0wCgYIKoZIzj0EAwIDRwAwRAIgK+Kv7G+/KkWkNZg3PcQFp866Z7G6soxo\\na4etSZ+SRlYCIBSiXS20Wc+yjD111nPzvQUCfsP4+DKZ3K+2GKsERD6d\\n-----END CERTIFICATE-----\\n\', md5=\'a07ee2076a6e392e1e96481e99ba094b\', sha1=\'8463b3a92912ccfd1d314705989bec139937d0d7\', sha256=\'0ae384bfd4dde9d13e50c5857c05a442c93f8e01445ee4b34540d22bd1e37f1b\', murmur=None)) ssh=None source=\'shodan\'\n\nService: port=80 banner=\'HTTP/1.1 301 Moved Permanently\\r\\nContent-Length: 0\\r\\nLocation: https://140.82.121.4/\\r\\n\\r\\n\' md5=\'86c9c13165fff2c41db667dd1f6500db\' sha1=\'8ecee0314c63c092d449b0d37d00f0ff62448dd1\' sha256=\'715fd6cec44f12bb5e59f57bf9a1691708a5a2d2ef253399358ac3ff88f47781\' murmur=\'1256792822\' first_seen=datetime.datetime(2021, 8, 20, 16, 10, 27, 656097) last_seen=datetime.datetime(2021, 8, 20, 16, 10, 27, 656099) timestamp=datetime.datetime(2021, 8, 16, 20, 42, 38, 12570) http=HTTPComponent(headers={\'Content-Length\': \'0\', \'Location\': \'https://140.82.121.4/\'}, content=HTTPComponentContent(raw=\'\', length=0, md5=\'d41d8cd98f00b204e9800998ecf8427e\', sha1=\'da39a3ee5e6b4b0d3255bfef95601890afd80709\', sha256=\'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855\', murmur=\'0\', favicon=None, robots_txt=None, security_txt=None)) tls=None ssh=None source=\'shodan\'\n\nService: port=9418 banner=\'0020ERR \\n  Repository not found.\' md5=\'fc79d10176a28462bc071909fb99d59e\' sha1=\'0df752dd6b3d3a461f30691ff89ee67ef3dd21b1\' sha256=\'5f355dafa7999d06265bb17e949b76eab283e9966ef8e27908c4e1058a891121\' murmur=\'-1636861384\' first_seen=datetime.datetime(2021, 8, 20, 16, 10, 27, 656097) last_seen=datetime.datetime(2021, 8, 20, 16, 10, 27, 656099) timestamp=datetime.datetime(2021, 8, 12, 2, 10, 13, 136596) http=None tls=None ssh=None source=\'shodan\'\n\nService: port=22 banner=\'SSH-2.0-babeld-968490c5\\nKey type: ssh-rsa\\nKey: AAAAB3NzaC1yc2EAAAABIwAAAQEAq2A7hRGmdnm9tUDbO9IDSwBK6TbQa+PXYPCPy6rbTrTtw7PH\\nkccKrpp0yVhp5HdEIcKr6pLlVDBfOLX9QUsyCOV0wzfjIJNlGEYsdlLJizHhbn2mUjvSAHQqZETY\\nP81eFzLQNnPHt4EVVUh7VfDESU84KezmD5QlWpXLmvU31/yMf+Se8xhHTvKSCZIFImWwoG6mbUoW\\nf9nzpIoaSjB+weqqUUmpaaasXVal72J+UX2B+2RPW3RcT0eOzQgqlJL3RKrTJvdsjE3JEAvGq3lG\\nHSZXy28G3skua2SmVi/w4yCE6gbODqnTWlg7+wC604ydGXA8VJiS5ap43JXiUFFAaQ==\\nFingerprint: 16:27:ac:a5:76:28:2d:36:63:1b:56:4d:eb:df:a6:48\\n\\nKex Algorithms:\\n\\tcurve25519-sha256\\n\\tcurve25519-sha256@libssh.org\\n\\tecdh-sha2-nistp256\\n\\tecdh-sha2-nistp384\\n\\tecdh-sha2-nistp521\\n\\tdiffie-hellman-group-exchange-sha256\\n\\nServer Host Key Algorithms:\\n\\trsa-sha2-512\\n\\trsa-sha2-256\\n\\tssh-rsa\\n\\tssh-dss\\n\\nEncryption Algorithms:\\n\\tchacha20-poly1305@openssh.com\\n\\taes256-gcm@openssh.com\\n\\taes128-gcm@openssh.com\\n\\taes256-ctr\\n\\taes192-ctr\\n\\taes128-ctr\\n\\taes256-cbc\\n\\taes192-cbc\\n\\taes128-cbc\\n\\nMAC Algorithms:\\n\\thmac-sha2-512-etm@openssh.com\\n\\thmac-sha2-256-etm@openssh.com\\n\\thmac-sha2-512\\n\\thmac-sha2-256\\n\\thmac-sha1-etm@openssh.com\\n\\thmac-sha1\\n\\nCompression Algorithms:\\n\\tnone\\n\\n\' md5=\'d652c84a47553c66047cf67e51f1345a\' sha1=\'679cc04ce70413f639351328bb2f0a70545d0cb6\' sha256=\'2431ffe5d8acbb77b6db91ee84f75baa0422d56947e0b85331dde2560e5454d5\' murmur=\'1023136053\' first_seen=datetime.datetime(2021, 8, 20, 16, 10, 27, 656097) last_seen=datetime.datetime(2021, 8, 20, 16, 10, 27, 656099) timestamp=datetime.datetime(2021, 8, 3, 21, 10, 29, 568093) http=None tls=None ssh=SSHComponent(algorithms=SSHComponentAlgorithms(encryption=[\'chacha20-poly1305@openssh.com\', \'aes256-gcm@openssh.com\', \'aes128-gcm@openssh.com\', \'aes256-ctr\', \'aes192-ctr\', \'aes128-ctr\', \'aes256-cbc\', \'aes192-cbc\', \'aes128-cbc\'], key_exchange=[\'curve25519-sha256\', \'curve25519-sha256@libssh.org\', \'ecdh-sha2-nistp256\', \'ecdh-sha2-nistp384\', \'ecdh-sha2-nistp521\', \'diffie-hellman-group-exchange-sha256\'], mac=[\'hmac-sha2-512-etm@openssh.com\', \'hmac-sha2-256-etm@openssh.com\', \'hmac-sha2-512\', \'hmac-sha2-256\', \'hmac-sha1-etm@openssh.com\', \'hmac-sha1\'], key_algorithms=[\'rsa-sha2-512\', \'rsa-sha2-256\', \'ssh-rsa\', \'ssh-dss\'], compression=[\'none\']), key=SSHComponentKey(raw=\'AAAAB3NzaC1yc2EAAAABIwAAAQEAq2A7hRGmdnm9tUDbO9IDSwBK6TbQa+PXYPCPy6rbTrTtw7PH\\nkccKrpp0yVhp5HdEIcKr6pLlVDBfOLX9QUsyCOV0wzfjIJNlGEYsdlLJizHhbn2mUjvSAHQqZETY\\nP81eFzLQNnPHt4EVVUh7VfDESU84KezmD5QlWpXLmvU31/yMf+Se8xhHTvKSCZIFImWwoG6mbUoW\\nf9nzpIoaSjB+weqqUUmpaaasXVal72J+UX2B+2RPW3RcT0eOzQgqlJL3RKrTJvdsjE3JEAvGq3lG\\nHSZXy28G3skua2SmVi/w4yCE6gbODqnTWlg7+wC604ydGXA8VJiS5ap43JXiUFFAaQ==\\n\', type=\'ssh-rsa\', md5=\'1627aca576282d36631b564debdfa648\', sha1=\'bf6b6825d2977c511a475bbefb88aad54a92ac73\', sha256=\'9d385b83a9175292561a5ec4d4818e0aca51a264f17420112ef88ac3a139498f\', murmur=\'-388505952\')) source=\'shodan\'\n\n--- flattened JSON dump ---\n{\n  "443.port": 443,\n  "443.banner": "HTTP/1.1 301 Moved Permanently\\r\\nContent-Length: 0\\r\\nLocation: https://github.com/\\r\\n\\r\\n",\n  "443.md5": "d402a6212741f3690b4fa1e46d9bd8b6",\n  "443.sha1": "a24eb4ba0332776d38050a7b41d0366742dbf262",\n  "443.sha256": "5fbe0315395986d131e4948888987319e88e3e1da6c5460e8a0bf8b7a1e639f0",\n  "443.murmur": "-1655207803",\n  "443.first_seen": "2021-08-20T16:10:27.656097",\n  "443.last_seen": "2021-08-20T16:10:27.656099",\n  "443.timestamp": "2021-08-16T20:42:40.325940",\n  "443.http.headers.Content-Length": "0",\n  "443.http.headers.Location": "https://github.com/",\n  "443.http.content.raw": "",\n  "443.http.content.length": 0,\n  "443.http.content.md5": "d41d8cd98f00b204e9800998ecf8427e",\n  "443.http.content.sha1": "da39a3ee5e6b4b0d3255bfef95601890afd80709",\n  "443.http.content.sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",\n  "443.http.content.murmur": "0",\n  "443.tls.certificate.issuer.dn": "C=US, O=DigiCert, Inc., CN=DigiCert High Assurance TLS Hybrid ECC SHA256 2020 CA1",\n  "443.tls.certificate.issuer.country": "US",\n  "443.tls.certificate.issuer.organization": "DigiCert, Inc.",\n  "443.tls.certificate.issuer.common_name": "DigiCert High Assurance TLS Hybrid ECC SHA256 2020 CA1",\n  "443.tls.certificate.subject.dn": "C=US, ST=California, L=San Francisco, O=GitHub, Inc., CN=github.com",\n  "443.tls.certificate.subject.country": "US",\n  "443.tls.certificate.subject.state": "California",\n  "443.tls.certificate.subject.locality": "San Francisco",\n  "443.tls.certificate.subject.organization": "GitHub, Inc.",\n  "443.tls.certificate.subject.common_name": "github.com",\n  "443.tls.certificate.issued": "2021-03-25T00:00:00",\n  "443.tls.certificate.expires": "2022-03-30T23:59:59",\n  "443.tls.certificate.expired": false,\n  "443.tls.certificate.alternative_names": [\n    "www.github.com",\n    "github.com"\n  ],\n  "443.tls.certificate.pem": "-----BEGIN CERTIFICATE-----\\nMIIFBjCCBK2gAwIBAgIQDovzdw2S0Zbwu2H5PEFmvjAKBggqhkjOPQQDAjBnMQsw\\nCQYDVQQGEwJVUzEXMBUGA1UEChMORGlnaUNlcnQsIEluYy4xPzA9BgNVBAMTNkRp\\nZ2lDZXJ0IEhpZ2ggQXNzdXJhbmNlIFRMUyBIeWJyaWQgRUNDIFNIQTI1NiAyMDIw\\nIENBMTAeFw0yMTAzMjUwMDAwMDBaFw0yMjAzMzAyMzU5NTlaMGYxCzAJBgNVBAYT\\nAlVTMRMwEQYDVQQIEwpDYWxpZm9ybmlhMRYwFAYDVQQHEw1TYW4gRnJhbmNpc2Nv\\nMRUwEwYDVQQKEwxHaXRIdWIsIEluYy4xEzARBgNVBAMTCmdpdGh1Yi5jb20wWTAT\\nBgcqhkjOPQIBBggqhkjOPQMBBwNCAASt9vd1sdNJVApdEHG93CUGSyIcoiNOn6H+\\nudCMvTm8DCPHz5GmkFrYRasDE77BI3q5xMidR/aW4Ll2a1A2ZvcNo4IDOjCCAzYw\\nHwYDVR0jBBgwFoAUUGGmoNI1xBEqII0fD6xC8M0pz0swHQYDVR0OBBYEFCexfp+7\\nJplQ2PPDU1v+MRawux5yMCUGA1UdEQQeMByCCmdpdGh1Yi5jb22CDnd3dy5naXRo\\ndWIuY29tMA4GA1UdDwEB/wQEAwIHgDAdBgNVHSUEFjAUBggrBgEFBQcDAQYIKwYB\\nBQUHAwIwgbEGA1UdHwSBqTCBpjBRoE+gTYZLaHR0cDovL2NybDMuZGlnaWNlcnQu\\nY29tL0RpZ2lDZXJ0SGlnaEFzc3VyYW5jZVRMU0h5YnJpZEVDQ1NIQTI1NjIwMjBD\\nQTEuY3JsMFGgT6BNhktodHRwOi8vY3JsNC5kaWdpY2VydC5jb20vRGlnaUNlcnRI\\naWdoQXNzdXJhbmNlVExTSHlicmlkRUNDU0hBMjU2MjAyMENBMS5jcmwwPgYDVR0g\\nBDcwNTAzBgZngQwBAgIwKTAnBggrBgEFBQcCARYbaHR0cDovL3d3dy5kaWdpY2Vy\\ndC5jb20vQ1BTMIGSBggrBgEFBQcBAQSBhTCBgjAkBggrBgEFBQcwAYYYaHR0cDov\\nL29jc3AuZGlnaWNlcnQuY29tMFoGCCsGAQUFBzAChk5odHRwOi8vY2FjZXJ0cy5k\\naWdpY2VydC5jb20vRGlnaUNlcnRIaWdoQXNzdXJhbmNlVExTSHlicmlkRUNDU0hB\\nMjU2MjAyMENBMS5jcnQwDAYDVR0TAQH/BAIwADCCAQUGCisGAQQB1nkCBAIEgfYE\\ngfMA8QB2ACl5vvCeOTkh8FZzn2Old+W+V32cYAr4+U1dJlwlXceEAAABeGq/vRoA\\nAAQDAEcwRQIhAJ7miER//DRFnDJNn6uUhgau3WMt4vVfY5dGigulOdjXAiBIVCfR\\nxjK1v4F31+sVaKzyyO7JAa0fzDQM7skQckSYWQB3ACJFRQdZVSRWlj+hL/H3bYbg\\nIyZjrcBLf13Gg1xu4g8CAAABeGq/vTkAAAQDAEgwRgIhAJgAEkoJQRivBlwo7x67\\n3oVsf1ip096WshZqmRCuL/JpAiEA3cX4rb3waLDLq4C48NSoUmcw56PwO/m2uwnQ\\nprb+yh0wCgYIKoZIzj0EAwIDRwAwRAIgK+Kv7G+/KkWkNZg3PcQFp866Z7G6soxo\\na4etSZ+SRlYCIBSiXS20Wc+yjD111nPzvQUCfsP4+DKZ3K+2GKsERD6d\\n-----END CERTIFICATE-----\\n",\n  "443.tls.certificate.md5": "a07ee2076a6e392e1e96481e99ba094b",\n  "443.tls.certificate.sha1": "8463b3a92912ccfd1d314705989bec139937d0d7",\n  "443.tls.certificate.sha256": "0ae384bfd4dde9d13e50c5857c05a442c93f8e01445ee4b34540d22bd1e37f1b",\n  "443.source": "shodan",\n  "80.port": 80,\n  "80.banner": "HTTP/1.1 301 Moved Permanently\\r\\nContent-Length: 0\\r\\nLocation: https://140.82.121.4/\\r\\n\\r\\n",\n  "80.md5": "86c9c13165fff2c41db667dd1f6500db",\n  "80.sha1": "8ecee0314c63c092d449b0d37d00f0ff62448dd1",\n  "80.sha256": "715fd6cec44f12bb5e59f57bf9a1691708a5a2d2ef253399358ac3ff88f47781",\n  "80.murmur": "1256792822",\n  "80.first_seen": "2021-08-20T16:10:27.656097",\n  "80.last_seen": "2021-08-20T16:10:27.656099",\n  "80.timestamp": "2021-08-16T20:42:38.012570",\n  "80.http.headers.Content-Length": "0",\n  "80.http.headers.Location": "https://140.82.121.4/",\n  "80.http.content.raw": "",\n  "80.http.content.length": 0,\n  "80.http.content.md5": "d41d8cd98f00b204e9800998ecf8427e",\n  "80.http.content.sha1": "da39a3ee5e6b4b0d3255bfef95601890afd80709",\n  "80.http.content.sha256": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",\n  "80.http.content.murmur": "0",\n  "80.source": "shodan",\n  "9418.port": 9418,\n  "9418.banner": "0020ERR \\n  Repository not found.",\n  "9418.md5": "fc79d10176a28462bc071909fb99d59e",\n  "9418.sha1": "0df752dd6b3d3a461f30691ff89ee67ef3dd21b1",\n  "9418.sha256": "5f355dafa7999d06265bb17e949b76eab283e9966ef8e27908c4e1058a891121",\n  "9418.murmur": "-1636861384",\n  "9418.first_seen": "2021-08-20T16:10:27.656097",\n  "9418.last_seen": "2021-08-20T16:10:27.656099",\n  "9418.timestamp": "2021-08-12T02:10:13.136596",\n  "9418.source": "shodan",\n  "22.port": 22,\n  "22.banner": "SSH-2.0-babeld-968490c5\\nKey type: ssh-rsa\\nKey: AAAAB3NzaC1yc2EAAAABIwAAAQEAq2A7hRGmdnm9tUDbO9IDSwBK6TbQa+PXYPCPy6rbTrTtw7PH\\nkccKrpp0yVhp5HdEIcKr6pLlVDBfOLX9QUsyCOV0wzfjIJNlGEYsdlLJizHhbn2mUjvSAHQqZETY\\nP81eFzLQNnPHt4EVVUh7VfDESU84KezmD5QlWpXLmvU31/yMf+Se8xhHTvKSCZIFImWwoG6mbUoW\\nf9nzpIoaSjB+weqqUUmpaaasXVal72J+UX2B+2RPW3RcT0eOzQgqlJL3RKrTJvdsjE3JEAvGq3lG\\nHSZXy28G3skua2SmVi/w4yCE6gbODqnTWlg7+wC604ydGXA8VJiS5ap43JXiUFFAaQ==\\nFingerprint: 16:27:ac:a5:76:28:2d:36:63:1b:56:4d:eb:df:a6:48\\n\\nKex Algorithms:\\n\\tcurve25519-sha256\\n\\tcurve25519-sha256@libssh.org\\n\\tecdh-sha2-nistp256\\n\\tecdh-sha2-nistp384\\n\\tecdh-sha2-nistp521\\n\\tdiffie-hellman-group-exchange-sha256\\n\\nServer Host Key Algorithms:\\n\\trsa-sha2-512\\n\\trsa-sha2-256\\n\\tssh-rsa\\n\\tssh-dss\\n\\nEncryption Algorithms:\\n\\tchacha20-poly1305@openssh.com\\n\\taes256-gcm@openssh.com\\n\\taes128-gcm@openssh.com\\n\\taes256-ctr\\n\\taes192-ctr\\n\\taes128-ctr\\n\\taes256-cbc\\n\\taes192-cbc\\n\\taes128-cbc\\n\\nMAC Algorithms:\\n\\thmac-sha2-512-etm@openssh.com\\n\\thmac-sha2-256-etm@openssh.com\\n\\thmac-sha2-512\\n\\thmac-sha2-256\\n\\thmac-sha1-etm@openssh.com\\n\\thmac-sha1\\n\\nCompression Algorithms:\\n\\tnone\\n\\n",\n  "22.md5": "d652c84a47553c66047cf67e51f1345a",\n  "22.sha1": "679cc04ce70413f639351328bb2f0a70545d0cb6",\n  "22.sha256": "2431ffe5d8acbb77b6db91ee84f75baa0422d56947e0b85331dde2560e5454d5",\n  "22.murmur": "1023136053",\n  "22.first_seen": "2021-08-20T16:10:27.656097",\n  "22.last_seen": "2021-08-20T16:10:27.656099",\n  "22.timestamp": "2021-08-03T21:10:29.568093",\n  "22.ssh.algorithms.encryption": [\n    "chacha20-poly1305@openssh.com",\n    "aes256-gcm@openssh.com",\n    "aes128-gcm@openssh.com",\n    "aes256-ctr",\n    "aes192-ctr",\n    "aes128-ctr",\n    "aes256-cbc",\n    "aes192-cbc",\n    "aes128-cbc"\n  ],\n  "22.ssh.algorithms.key_exchange": [\n    "curve25519-sha256",\n    "curve25519-sha256@libssh.org",\n    "ecdh-sha2-nistp256",\n    "ecdh-sha2-nistp384",\n    "ecdh-sha2-nistp521",\n    "diffie-hellman-group-exchange-sha256"\n  ],\n  "22.ssh.algorithms.mac": [\n    "hmac-sha2-512-etm@openssh.com",\n    "hmac-sha2-256-etm@openssh.com",\n    "hmac-sha2-512",\n    "hmac-sha2-256",\n    "hmac-sha1-etm@openssh.com",\n    "hmac-sha1"\n  ],\n  "22.ssh.algorithms.key_algorithms": [\n    "rsa-sha2-512",\n    "rsa-sha2-256",\n    "ssh-rsa",\n    "ssh-dss"\n  ],\n  "22.ssh.algorithms.compression": [\n    "none"\n  ],\n  "22.ssh.key.raw": "AAAAB3NzaC1yc2EAAAABIwAAAQEAq2A7hRGmdnm9tUDbO9IDSwBK6TbQa+PXYPCPy6rbTrTtw7PH\\nkccKrpp0yVhp5HdEIcKr6pLlVDBfOLX9QUsyCOV0wzfjIJNlGEYsdlLJizHhbn2mUjvSAHQqZETY\\nP81eFzLQNnPHt4EVVUh7VfDESU84KezmD5QlWpXLmvU31/yMf+Se8xhHTvKSCZIFImWwoG6mbUoW\\nf9nzpIoaSjB+weqqUUmpaaasXVal72J+UX2B+2RPW3RcT0eOzQgqlJL3RKrTJvdsjE3JEAvGq3lG\\nHSZXy28G3skua2SmVi/w4yCE6gbODqnTWlg7+wC604ydGXA8VJiS5ap43JXiUFFAaQ==\\n",\n  "22.ssh.key.type": "ssh-rsa",\n  "22.ssh.key.md5": "1627aca576282d36631b564debdfa648",\n  "22.ssh.key.sha1": "bf6b6825d2977c511a475bbefb88aad54a92ac73",\n  "22.ssh.key.sha256": "9d385b83a9175292561a5ec4d4818e0aca51a264f17420112ef88ac3a139498f",\n  "22.ssh.key.murmur": "-388505952",\n  "22.source": "shodan"\n}\n```',
    'author': '3c7',
    'author_email': '3c7@posteo.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/3c7/common-osint-model',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from common_osint_model.shodan import from_shodan, from_shodan_flattened
from common_osint_model.censys.v1 import (
    from_censys_ipv4,
    from_censys_ipv4_flattened,
    from_censys_certificates,
    from_censys_certificates_flattened
)
from common_osint_model.censys.v2 import from_censys, from_censys_flattened
from common_osint_model.certificate import from_x509_pem, from_x509_pem_flattened

from common_osint_model.models.host import Host
from common_osint_model.models.domain import *
from common_osint_model.models.service import Service
from common_osint_model.models.autonomous_system import AutonomousSystem
from common_osint_model.models.http import *
from common_osint_model.models.tls import *
from common_osint_model.models.ssh import *



import io
import json
from argparse import ArgumentParser

from common_osint_model import *


def convshodan() -> None:
    """
    Converts a JSON file with shodan data to the common data model
    """
    args = parse_args()
    if args.flatten:
        convert(args.filepath, from_shodan_flattened, args.indent)
    else:
        convert(args.filepath, from_shodan, args.indent)


def convcensys():
    """
    Converts a JSON file with shodan data to the common data model
    """
    args = parse_args()
    if args.flatten:
        convert(args.filepath, from_censys_ipv4_flattened, args.indent)
    else:
        convert(args.filepath, from_censys_ipv4, args.indent)


def convcensyscert():
    """
    Converts a JSON file with censys certificate data to the common data model
    """
    args = parse_args()
    if args.flatten:
        convert(args.filepath, from_censys_certificates_flattened, args.indent)
    else:
        convert(args.filepath, from_censys_certificates, args.indent)


def convcensys2():
    """
    Converts a JSON file with censys search 2.0 data to the common data model
    """
    args = parse_args()
    if args.flatten:
        convert(args.filepath, from_censys_flattened, args.indent)
    else:
        convert(args.filepath, from_censys, args.indent)


def convcert():
    """
    Converts a certificate PEM file into the common data model
    """
    args = parse_args()
    if args.flatten:
        d = from_x509_pem_flattened(open(args.filepath).read())
    else:
        d = from_x509_pem(open(args.filepath).read())
    print(json.dumps(d, indent=args.indent))


def parse_args():
    parser = ArgumentParser()
    parser.add_argument("filepath", type=str, help="JSON file to read from")
    parser.add_argument(
        "-F", "--flatten", action="store_true", help="Flatten the output"
    )
    parser.add_argument(
        "-I", "--indent", type=int, default=4, help="Indent of the JSON output"
    )
    return parser.parse_args()


def convert(filepath: str, f, indent: int = 4) -> None:
    with io.open(filepath, "r") as file:
        data = json.loads(file.read())
        print(json.dumps(f(data), indent=indent))



from datetime import datetime
from typing import Optional, List

from pydantic import BaseModel


class Entity(BaseModel):
    """Represents an entity which registered a domain."""
    name: Optional[str]
    email: Optional[str]
    organization: Optional[str]
    street: Optional[str]
    city: Optional[str]
    state: Optional[str]
    postal_code: Optional[str]
    country: Optional[str]
    phone: Optional[str]
    timestamp: datetime = datetime.utcnow()


class Domain(BaseModel):
    """Represents a domain pointing to a specific host. Also, this object might be used to represent found via other
    sources, therefore a 'query' field might contain the query used to find it"""
    domain: str
    first_seen: datetime = datetime.utcnow()
    last_seen: datetime = datetime.utcnow()
    source: Optional[str]
    type: Optional[str]
    soa: Optional[List[str]]
    nameserver: Optional[List[str]]
    registrar: Optional[str]
    registrant: Optional[Entity]
    query: Optional[str]
